"""
    nis
    ===

    Free functions to abstractly wrap the NIS REST API with an abstract client.

    These functions return HTTP responses, or futures to responses, and do
    not do any formatting for the NEM response.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from . import client
from .. import util
from .. import models


OptionalNetworkType = typing.Optional[models.NetworkType]

# BOILERPLATE
# -----------


def synchronous_request(name, doc="", raise_for_status=True):
    """Generate wrappers for a synchronous request."""

    def f(client, network_type, *args, **kwds):
        request, process = CLIENT_CB[name]
        response = request(client, *args, **kwds)
        if raise_for_status:
            response.raise_for_status()
        status = response.status_code
        json = response.json()
        return process(status, json, network_type)

    f.__name__ = name
    f.__doc__ = doc
    f.func_name = name

    return f


def asynchronous_request(name, doc="", raise_for_status=True):
    """Generate wrappers for an asynchronous request."""

    async def f(client, network_awaitable, *args, **kwds):
        # Await the network type so if an exception is thrown, we
        # don't forget to await the awaitable.
        request, process = CLIENT_CB[name]
        network_type = await network_awaitable
        async with request(client, *args, **kwds) as response:
            if raise_for_status:
                response.raise_for_status()
            status = response.status
            json = await response.json()
            return process(status, json, network_type)

    f.__name__ = f"async_{name}"
    f.__doc__ = doc
    f.func_name = f"async_{name}"

    return f


def request(*args, **kwds):
    """Generate synchronous and asynchronous request wrappers."""

    s = synchronous_request(*args, **kwds)
    a = asynchronous_request(*args, **kwds)
    return s, a


# ACCOUNT HTTP
# ------------


def request_get_account_info(
    client: client.Client,
    address: models.Address,
    **kwds
):
    """
    Make "/account/{address}" request.

    :param client: Wrapper for client.
    :param address: Account address.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{address.address}"
    return client.get(url, **kwds)


def process_get_account_info(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.AccountInfo:
    """
    Process the "/account/{address}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.AccountInfo.create_from_dto(json, network_type)


get_account_info = request("get_account_info")


def request_get_accounts_info(
    client: client.Client,
    addresses: typing.Sequence[models.Address],
    **kwds
):
    """
    Make "/account" request.

    :param client: Wrapper for client.
    :param addresses: Sequence of account addresses.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/account"
    json = {"addresses": [i.address for i in addresses]}
    return client.post(url, json=json, **kwds)


def process_get_accounts_info(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.AccountInfo]:
    """
    Process the "/account" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return [models.AccountInfo.create_from_dto(i, network_type) for i in json]


get_accounts_info = request("get_accounts_info")


def request_get_account_properties(
    client: client.Client,
    address: models.Address,
    **kwds
):
    """
    Make "/account/{address}/properties" request.

    :param client: Wrapper for client.
    :param address: Account address.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{address.address}/properties"
    return client.get(url, **kwds)


def process_get_account_properties(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.AccountProperties:
    """
    Process the "/account/{address}/properties" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.AccountProperties.create_from_dto(json, network_type)


get_account_properties = request("get_account_properties")


def request_get_accounts_properties(
    client: client.Client,
    addresses: typing.Sequence[models.Address],
    **kwds
):
    """
    Make "/account/properties" request.

    :param client: Wrapper for client.
    :param addresses: Sequence of account addresses.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/account/properties"
    json = {"addresses": [i.address for i in addresses]}
    return client.post(url, json=json, **kwds)


def process_get_accounts_properties(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> typing.Sequence[models.AccountProperties]:
    """
    Process the "/account/properties" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return [models.AccountProperties.create_from_dto(i, network_type) for i in json]


get_accounts_properties = request("get_accounts_properties")


def request_get_multisig_account_info(
    client: client.Client,
    address: models.Address,
    **kwds
):
    """
    Make "/account/{address}/multisig" request.

    :param client: Wrapper for client.
    :param address: Account address.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{address.address}/multisig"
    return client.get(url, **kwds)


def process_get_multisig_account_info(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.MultisigAccountInfo:
    """
    Process the "/account/{address}/multisig" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.MultisigAccountInfo.create_from_dto(json, network_type)


get_multisig_account_info = request("get_multisig_account_info")


def request_get_multisig_account_graph_info(
    client: client.Client,
    address: models.Address,
    **kwds
):
    """
    Make "/account/{address}/multisig/graph" request.

    :param client: Wrapper for client.
    :param address: Account address.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{address.address}/multisig/graph"
    return client.get(url, **kwds)


def process_get_multisig_account_graph_info(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> models.MultisigAccountInfo:
    """
    Process the "/account/{address}/multisig/graph" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.MultisigAccountGraphInfo.create_from_dto(json, network_type)


get_multisig_account_graph_info = request("get_multisig_account_graph_info")


def request_get_account_transactions(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/transactions" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/transactions"
    return client.get(url, **kwds)


def process_get_account_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/account/{public_key}/transactions" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_account_transactions = request("get_account_transactions")


def request_get_account_incoming_transactions(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/transactions/incoming" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/transactions/incoming"
    return client.get(url, **kwds)


def process_get_account_incoming_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/account/{public_key}/transactions/incoming" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_account_incoming_transactions = request("get_account_incoming_transactions")


def request_get_account_outgoing_transactions(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/transactions/outgoing" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/transactions/outgoing"
    return client.get(url, **kwds)


def process_get_account_outgoing_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/account/{public_key}/transactions/outgoing" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_account_outgoing_transactions = request("get_account_outgoing_transactions")


def request_get_account_unconfirmed_transactions(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/transactions/unconfirmed" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/transactions/unconfirmed"
    return client.get(url, **kwds)


def process_get_account_unconfirmed_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/account/{public_key}/transactions/unconfirmed" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_account_unconfirmed_transactions = request("get_account_unconfirmed_transactions")


def request_get_account_partial_transactions(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/transactions/partial" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/transactions/partial"
    return client.get(url, **kwds)


def process_get_account_partial_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/account/{public_key}/transactions/partial" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_account_partial_transactions = request("get_account_partial_transactions")
# TODO: Check when stabilized
#
#
# def request_get_account_contracts(
#     client: client.Client,
#     public_account: models.PublicAccount,
#     **kwds
# ):
#     """
#     Make "/account/{public_key}/contracts" request.
#
#     :param client: Wrapper for client.
#     :param public_account: Public account.
#     :param timeout: (Optional) timeout for request (in seconds).
#     """
#
#     url = f"/account/{public_account.public_key}/contracts"
#     return client.get(url, **kwds)
#
#
# def process_get_account_contracts(
#     status: int,
#     json: list,
#     network_type: models.NetworkType,
# ) -> typing.Sequence[models.Transaction]:
#     """
#     Process the "/account/{public_key}/contracts" HTTP response.
#
#     :param status: Status code for HTTP response.
#     :param json: JSON data for response message.
#     :param network_type: Network type..
#     """
#
#     assert status == 200
#     return [models.ContractInfo.create_from_dto(i, network_type) for i in json]
#
#
# get_account_contracts = request("get_account_contracts")


def request_get_account_names(
    client: client.Client,
    addresses: typing.Sequence[models.Address],
    **kwds
):
    """
    Make "/account/names" request.

    :param client: Wrapper for client.
    :param addresses: Sequence of account addresses.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/account/names"
    json = {"addresses": [i.address for i in addresses]}
    return client.post(url, json=json, **kwds)


def process_get_account_names(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> typing.Sequence[models.AccountNames]:
    """
    Process the "/account/names" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return [models.AccountNames.create_from_dto(i, network_type) for i in json]


get_account_names = request("get_account_names")


# BLOCKCHAIN HTTP
# ---------------


def request_get_block_by_height(
    client: client.Client,
    height: int,
    **kwds
):
    """
    Make "/block/{height}" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/block/{height}"
    return client.get(url, **kwds)


def process_get_block_by_height(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.BlockInfo:
    """
    Process the "/block/{height}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return models.BlockInfo.create_from_dto(json, network_type)


get_block_by_height = request("get_block_by_height")


def request_get_blocks_by_height_with_limit(
    client: client.Client,
    height: int,
    limit: int,
    **kwds
):
    """
    Make "/blocks/{height}/limit/{limit}" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param limit: Maximum number of blocks to return.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/blocks/{height}/limit/{limit}"
    return client.get(url, **kwds)


def process_get_blocks_by_height_with_limit(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.BlockInfo]:
    """
    Process the "/blocks/{height}/limit/{limit}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.BlockInfo.create_from_dto(i, network_type) for i in json]


get_blocks_by_height_with_limit = request("get_blocks_by_height_with_limit")


def request_get_block_transactions(
    client: client.Client,
    height: int,
    **kwds
):
    """
    Make "/block/{height}/transactions" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/block/{height}/transactions"
    return client.get(url, **kwds)


def process_get_block_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.Transaction]:
    """
    Process the "/block/{height}/transactions" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_block_transactions = request("get_block_transactions")


def request_get_merkle_by_hash_in_block(
    client: client.Client,
    height: int,
    hash: str,
    **kwds
):
    """
    Make "/block/{height}/transaction/{hash}/merkle" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param hash: Transaction hash included in block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/block/{height}/transaction/{hash}/merkle"
    return client.get(url, **kwds)


def process_get_merkle_by_hash_in_block(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.MerkleProofInfo:
    """
    Process the "/block/{height}/transaction/{hash}/merkle" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return models.MerkleProofInfo.create_from_dto(json, network_type)


get_merkle_by_hash_in_block = request("get_merkle_by_hash_in_block")


def request_get_block_receipts(
    client: client.Client,
    height: int,
    **kwds
):
    """
    Make "/block/{height}/receipts" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/block/{height}/receipts"
    return client.get(url, **kwds)


def process_get_block_receipts(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.Statements:
    """
    Process the "/block/{height}/receipts" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return models.Statements.create_from_dto(json, network_type)


get_block_receipts = request("get_block_receipts")


def request_get_blockchain_height(client: client.Client, **kwds):
    """
    Make "/chain/height" request.

    :param client: Wrapper for client.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/chain/height"
    return client.get(url, **kwds)


def process_get_blockchain_height(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> int:
    """
    Process the "/chain/height" HTTP response.

    Note: The data-transfer object format for the `HeightDTO` is in
    the description for `BlockInfo`.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return util.u64_from_dto(json['height'])


get_blockchain_height = request("get_blockchain_height")


def request_get_blockchain_score(client: client.Client, **kwds):
    """
    Make "/chain/score" request.

    :param client: Wrapper for client.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/chain/score"
    return client.get(url, **kwds)


def process_get_blockchain_score(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.BlockchainScore:
    """
    Process the "/chain/score" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.BlockchainScore.create_from_dto(json, network_type)


get_blockchain_score = request("get_blockchain_score")


def request_get_diagnostic_blocks_by_height_with_limit(
    client: client.Client,
    height: int,
    limit: int,
    **kwds
):
    """
    Make "/diagnostic/blocks/{height}/limit/{limit}" request.

    :param client: Wrapper for client.
    :param height: Height of block.
    :param limit: Maximum number of blocks to return.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/diagnostic/blocks/{height}/limit/{limit}"
    return client.get(url, **kwds)


def process_get_diagnostic_blocks_by_height_with_limit(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.BlockInfo]:
    """
    Process the "/diagnostic/blocks/{height}/limit/{limit}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return [models.BlockInfo.create_from_dto(i, network_type) for i in json]


get_diagnostic_blocks_by_height_with_limit = request(
    "get_diagnostic_blocks_by_height_with_limit"
)


def request_get_diagnostic_storage(client: client.Client, **kwds):
    """
    Make "/diagnostic/storage" request.

    :param client: Wrapper for client.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/diagnostic/storage"
    return client.get(url, **kwds)


def process_get_diagnostic_storage(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.BlockchainStorageInfo:
    """
    Process the "/diagnostic/storage" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.BlockchainStorageInfo.create_from_dto(json, network_type)


get_diagnostic_storage = request("get_diagnostic_storage")


def request_get_diagnostic_server(client: client.Client, **kwds):
    """
    Make "/diagnostic/server" request.

    :param client: Wrapper for client.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/diagnostic/server"
    return client.get(url, **kwds)


def process_get_diagnostic_server(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.BlockchainServerInfo:
    """
    Process the "/diagnostic/server" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.BlockchainServerInfo.create_from_dto(json, network_type)


get_diagnostic_server = request("get_diagnostic_server")


# TODO: Check when stabilized
# CONTRACT HTTP
# -----------
#
#
# def request_get_contracts(
#     client: client.Client,
#     addresses: typing.Sequence[models.Address],
#     **kwds
# ):
#     """
#     Make "/contract" request.
#
#     :param addresses: Sequence of account addresses.
#     """
#
#     url = f"/contract"
#     json = {"addresses": [i.address for i in addresses]}
#     return client.post(url, json=json, **kwds)
#
#
# def process_get_contracts(
#     status: int,
#     json: list,
#     network_type: models.NetworkType,
# ) -> typing.Sequence[models.ContractInfo]:
#     """
#     Process the "/contract" HTTP response.
#
#     :param status: Status code for HTTP response.
#     :param json: JSON data for response message.
#     :param network_type: Network type.
#     """
#
#     assert status == 200
#     return [models.ContractInfo.create_from_dto(i, network_type) for i in json]
#
#
# get_contracts = request("get_contracts")
#
#
# def request_get_contract(
#     client: client.Client,
#     contract_id: str,
#     **kwds
# ):
#     """
#     Make "/contract/{contract_id}" request.
#
#     :param contract_id: The account identifier.
#     """
#
#     url = f"/contract/{contract_id}"
#     return client.get(url, **kwds)
#
#
# def process_get_contract(
#     status: int,
#     json: dict,
#     network_type: models.NetworkType,
# ) -> models.ContractInfo:
#     """
#     Process the "/contract/{contract_id}" HTTP response.
#
#     :param status: Status code for HTTP response.
#     :param json: JSON data for response message.
#     :param network_type: Network type.
#     """
#
#     assert status == 200
#     return models.ContractInfo.create_from_dto(json, network_type)
#
#
# get_contract = request("get_contract")
#
#
# METADATA HTTP
# -----------


def request_get_account_metadata(
    client: client.Client,
    public_account: models.PublicAccount,
    **kwds
):
    """
    Make "/account/{public_key}/metadata" request.

    :param client: Wrapper for client.
    :param public_account: Public account.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{public_account.public_key}/metadata"
    return client.get(url, **kwds)


def process_get_account_metadata(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.AddressMetadataInfo:
    """
    Process the "/account/{public_key}/metadata" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return models.AddressMetadataInfo.create_from_dto(json)


get_account_metadata = request("get_account_metadata")


def request_get_mosaic_metadata(
    client: client.Client,
    mosaic_id: models.MosaicId,
    **kwds
):
    """
    Make "/mosaic/{mosaic_id}/metadata" request.

    :param client: Wrapper for client.
    :param mosaic_id: Mosaic ID.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/mosaic/{mosaic_id:016x}/metadata"
    return client.get(url, **kwds)


def process_get_mosaic_metadata(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.MosaicMetadataInfo:
    """
    Process the "/mosaic/{mosaic_id}/metadata" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type..
    """

    assert status == 200
    return models.MosaicMetadataInfo.create_from_dto(json)


get_mosaic_metadata = request("get_mosaic_metadata")


def request_get_namespace_metadata(
    client: client.Client,
    namespace_id: models.NamespaceId,
    **kwds
):
    """
    Make "/namespace/{namespace_id}/metadata" request.

    :param client: Wrapper for client.
    :param id: Namespace ID.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/namespace/{namespace_id:016x}/metadata"
    return client.get(url, **kwds)


def process_get_namespace_metadata(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.NamespaceMetadataInfo:
    """
    Process the "/namespace/{namespace_id}/metadata" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.NamespaceMetadataInfo.create_from_dto(json, network_type)


get_namespace_metadata = request("get_namespace_metadata")


def request_get_metadata(
    client: client.Client,
    metadata_id: str,
    **kwds
):
    """
    Make "/metadata/{metadata_id}" request.

    :param metadata_id: ID of metadata.
    """

    url = f"/metadata/{metadata_id}"
    return client.get(url, **kwds)


def process_get_metadata(
    status: int,
    json: dict,
    network_type: models.NetworkType,
):
    """
    Process the "/metadata/{metadata_id}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.MetadataInfo.create_from_dto(json)


get_metadata = request("get_metadata")


def request_get_metadatas(
    client: client.Client,
    metadata_ids: typing.Sequence[str],
    **kwds
):
    """
    Make "/metadata" request.

    """

    url = "/metadata"
    json = {"metadataIds": metadata_ids}
    return client.post(url, json=json, **kwds)


def process_get_metadatas(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.MetadataInfo]:
    """
    Process the "/metadata" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return [models.MetadataInfo.create_from_dto(i, network_type) for i in json]


get_metadatas = request("get_metadatas")


# CONFIG HTTP
# -----------


def request_get_config(
    client: client.Client,
    height: int,
    **kwds
):
    """
    Make "/config/{height}" request.

    :param height: The height of the blockchain to get config.
    """

    url = f"/config/{height}"
    return client.get(url, **kwds)


def process_get_config(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.CatapultConfig:
    """
    Process the "/config/{height}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.CatapultConfig.create_from_dto(json, network_type)


get_config = request("get_config")


def request_get_upgrade(
    client: client.Client,
    height: int,
    **kwds
):
    """
    Make "/upgrade/{height}" request.

    :param height: The height of the blockchain to get upgrade.
    """

    url = f"/upgrade/{height}"
    return client.get(url, **kwds)


def process_get_upgrade(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.CatapultUpgrade:
    """
    Process the "/upgrade/{height}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.CatapultUpgrade.create_from_dto(json, network_type)


get_upgrade = request("get_upgrade")


# NODE HTTP
# -----------


def request_get_node_info(
    client: client.Client,
    **kwds
):
    """
    Make "/node/info" request.

    """

    url = "/node/info"
    return client.get(url, **kwds)


def process_get_node_info(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.NodeInfo:
    """
    Process the "/node/info" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.NodeInfo.create_from_dto(json, network_type)


get_node_info = request("get_node_info")


def request_get_node_time(
    client: client.Client,
    **kwds
):
    """
    Make "/node/time" request.

    """

    url = "/node/time"
    return client.get(url, **kwds)


def process_get_node_time(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.NodeTime:
    """
    Process the "/node/time" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    :param network_type: Network type.
    """

    assert status == 200
    return models.NodeTime.create_from_dto(json, network_type)


get_node_time = request("get_node_time")


# MOSAIC HTTP
# -----------


def request_get_mosaic(
    client: client.Client,
    id: models.MosaicId,
    **kwds
):
    """
    Make "/mosaic/{id}" request.

    :param client: Wrapper for client.
    :param id: Mosaic ID to request info for.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/mosaic/{id:016x}"
    return client.get(url, **kwds)


def process_get_mosaic(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> typing.Sequence[models.MosaicName]:
    """
    Process the "/mosaic/{id}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.MosaicInfo.create_from_dto(json, network_type)


get_mosaic = request("get_mosaic")


def request_get_mosaics(
    client: client.Client,
    ids: typing.Sequence[models.MosaicId],
    **kwds
):
    """
    Make "/mosaic" request.

    :param client: Wrapper for client.
    :param ids: Mosaic IDs to request names for.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/mosaic"
    json = {"mosaicIds": [f"{i:016x}" for i in ids]}
    return client.post(url, json=json, **kwds)


def process_get_mosaics(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.MosaicName]:
    """
    Process the "/mosaic" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.MosaicInfo.create_from_dto(i, network_type) for i in json]


get_mosaics = request("get_mosaics")


def request_get_mosaic_names(
    client: client.Client,
    ids: typing.Sequence[models.MosaicId],
    **kwds
):
    """
    Make "/mosaic/names" request.

    :param client: Wrapper for client.
    :param ids: Mosaic IDs to request names for.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/mosaic/names"
    json = {"mosaicIds": [f"{i:016x}" for i in ids]}
    return client.post(url, json=json, **kwds)


def process_get_mosaic_names(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.MosaicName]:
    """
    Process the "/mosaic/names" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.MosaicName.create_from_dto(i, network_type) for i in json]


get_mosaic_names = request("get_mosaic_names")


def request_get_mosaic_richlist(
    client: client.Client,
    mosaic_id: str,
    **kwds
):
    """
    Make "/mosaic/{mosaic_id}/richlist" request.

    :param client: Wrapper for client.
    :param mosaic_id: Mosaic IDs to request richlist for.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/mosaic/{mosaic_id}/richlist"
    return client.get(url, **kwds)


def process_get_mosaic_richlist(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.AccountBalance]:
    """
    Process the "/mosaic/{mosaic_id}/richlist" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.AccountBalance.create_from_dto(i, network_type) for i in json]


get_mosaic_richlist = request("get_mosaic_richlist")

# NAMESPACE HTTP
# --------------


def request_get_namespace(
    client: client.Client,
    namespace_id: models.NamespaceId,
    **kwds
):
    """
    Make "/namespace/{namespace_id}" request.

    :param client: Wrapper for client.
    :param id: Namespace ID.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/namespace/{namespace_id:016x}"
    return client.get(url, **kwds)


def process_get_namespace(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.NamespaceInfo:
    """
    Process the "/namespace/{namespace_id}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.NamespaceInfo.create_from_dto(json, network_type)


get_namespace = request("get_namespace")


def request_get_namespaces_name(
    client: client.Client,
    ids: typing.Sequence[models.NamespaceId],
    **kwds
):
    """
    Make "/namespace/names" request.

    :param client: Wrapper for client.
    :param ids: Namespace IDs to request names for.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/namespace/names"
    json = {"namespaceIds": [f"{i:016x}" for i in ids]}
    return client.post(url, json=json, **kwds)


def process_get_namespaces_name(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.NamespaceName]:
    """
    Process the "/namespace/names" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.NamespaceName.create_from_dto(i, network_type) for i in json]


get_namespaces_name = request("get_namespaces_name")


def request_get_namespaces_from_account(
    client: client.Client,
    address: models.Address,
    **kwds
):
    """
    Make "/account/{address}/namespaces" request.

    :param client: Wrapper for client.
    :param address: Account address.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/account/{address.address}/namespaces"
    return client.get(url, **kwds)


def process_get_namespaces_from_account(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.NamespaceInfo]:
    """
    Process the "/account/{address}/namespaces" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.NamespaceInfo.create_from_dto(i, network_type) for i in json]


get_namespaces_from_account = request("get_namespaces_from_account")


def request_get_namespaces_from_accounts(
    client: client.Client,
    addresses: typing.Sequence[models.Address],
    **kwds
):
    """
    Make "/account/namespaces" request.

    :param client: Wrapper for client.
    :param addresses: Sequence of account addresses.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/account/namespaces"
    json = {"addresses": [i.address for i in addresses]}
    return client.post(url, json=json, **kwds)


def process_get_namespaces_from_accounts(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.NamespaceInfo]:
    """
    Process the "/account/namespaces" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.NamespaceInfo.create_from_dto(i, network_type) for i in json]


get_namespaces_from_accounts = request("get_namespaces_from_accounts")


def request_get_linked_mosaic_id(
    client: client.Client,
    namespace_id: models.NamespaceId,
    **kwds
):
    """
    Make "/namespace/{namespace_id}" request.

    :param client: Wrapper for client.
    :param id: Namespace ID.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    return request_get_namespace(client, namespace_id, **kwds)


def process_get_linked_mosaic_id(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.MosaicId:
    """
    Process the "/namespace/{namespace_id}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    namespace_info = process_get_namespace(status, json, network_type)
    return namespace_info.alias.mosaic_id


get_linked_mosaic_id = request("get_linked_mosaic_id")


def request_get_linked_address(
    client: client.Client,
    namespace_id: models.NamespaceId,
    **kwds
):
    """
    Make "/namespace/{namespace_id}" request.

    :param client: Wrapper for client.
    :param id: Namespace ID.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    return request_get_namespace(client, namespace_id, **kwds)


def process_get_linked_address(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.Address:
    """
    Process the "/namespace/{namespace_id}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    namespace_info = process_get_namespace(status, json, network_type)
    return namespace_info.alias.address


get_linked_address = request("get_linked_address")


# NETWORK HTTP
# ------------

NETWORK_TYPE = {
    'mijin': models.NetworkType.MIJIN,
    'mijinTest': models.NetworkType.MIJIN_TEST,
    'public': models.NetworkType.MAIN_NET,
    'publicTest': models.NetworkType.TEST_NET,
}


def request_get_network_type(client: client.Client, **kwds):
    """
    Make "/network" request.

    Note: The data-transfer object format for the `NetworkTypeDTO` is in
    the description for `NetworkType`.

    :param client: Wrapper for client.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/network"
    return client.get(url, **kwds)


def process_get_network_type(
    status: int,
    json: dict,
    network_type: OptionalNetworkType,
) -> models.NetworkType:
    """
    Process the "/network" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return NETWORK_TYPE[json['name']]


get_network_type = request("get_network_type")

# TRANSACTION HTTP
# ----------------


def request_get_transaction(client: client.Client, hash: str, **kwds):
    """
    Make "/transaction/{hash}" request.

    :param client: Wrapper for client.
    :param hash: Transaction hash.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/transaction/{hash}"
    return client.get(url, **kwds)


def process_get_transaction(
    status: int,
    json: dict,
    network_type: models.NetworkType,
):
    """
    Process the "/transaction/{hash}" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.Transaction.create_from_dto(json, network_type)


get_transaction = request("get_transaction")


def request_get_transactions(
    client: client.Client,
    hashes: typing.Sequence[str],
    **kwds
):
    """
    Make "/transaction" request.

    :param client: Wrapper for client.
    :param hashes: Sequence of transaction hashes.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/transaction"
    json = {'transactionIds': list(hashes)}
    return client.post(url, json=json, **kwds)


def process_get_transactions(
    status: int,
    json: list,
    network_type: models.NetworkType,
):
    """
    Process the "/transaction" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.Transaction.create_from_dto(i, network_type) for i in json]


get_transactions = request("get_transactions")


def request_get_transaction_status(
    client: client.Client,
    hash: str,
    **kwds
):
    """
    Make "/transaction/{hash}/status" request.

    :param client: Wrapper for client.
    :param hash: Transaction hash.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = f"/transaction/{hash}/status"
    return client.get(url, **kwds)


def process_get_transaction_status(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.TransactionStatus:
    """
    Process the "/transaction/{hash}/status" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return models.TransactionStatus.create_from_dto(json, network_type)


get_transaction_status = request("get_transaction_status")


def request_get_transaction_statuses(
    client: client.Client,
    hashes: typing.Sequence[str],
    **kwds
):
    """
    Make "/transaction/statuses" request.

    :param client: Wrapper for client.
    :param hashes: Sequence of transaction hashes.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/transaction/statuses"
    json = {'hashes': list(hashes)}
    return client.post(url, json=json, **kwds)


def process_get_transaction_statuses(
    status: int,
    json: list,
    network_type: models.NetworkType,
) -> typing.Sequence[models.TransactionStatus]:
    """
    Process the "/transaction/statuses" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert status == 200
    return [models.TransactionStatus.create_from_dto(i, network_type) for i in json]


get_transaction_statuses = request("get_transaction_statuses")


def request_announce(
    client: client.Client,
    transaction: models.SignedTransaction,
    **kwds
):
    """
    Make "/transaction" request.

    :param client: Wrapper for client.
    :param transaction: Signed transaction data.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/transaction"
    json = {'payload': transaction.payload}
    return client.put(url, json=json, **kwds)


def process_announce(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.TransactionAnnounceResponse:
    """
    Process the "/transaction" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert ((status == 200) | (status == 202))
    return models.TransactionAnnounceResponse.create_from_dto(json)


announce = request("announce")


def request_announce_partial(
    client: client.Client,
    transaction: models.SignedTransaction,
    **kwds
):
    """
    Make "/transaction/partial" request.

    :param client: Wrapper for client.
    :param transaction: Signed transaction data.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/transaction/partial"
    json = {'payload': transaction.payload}
    return client.put(url, json=json, **kwds)


def process_announce_partial(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.TransactionAnnounceResponse:
    """
    Process the "/transaction/partial" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert ((status == 200) | (status == 202))
    return models.TransactionAnnounceResponse.create_from_dto(json)


announce_partial = request("announce_partial")


def request_announce_cosignature(
    client: client.Client,
    transaction: models.CosignatureSignedTransaction,
    **kwds
):
    """
    Make "/transaction/cosignature" request.

    :param client: Wrapper for client.
    :param transaction: Signed transaction data.
    :param timeout: (Optional) timeout for request (in seconds).
    """

    url = "/transaction/cosignature"
    json = transaction.to_dto()
    return client.put(url, json=json, **kwds)


def process_announce_cosignature(
    status: int,
    json: dict,
    network_type: models.NetworkType,
) -> models.TransactionAnnounceResponse:
    """
    Process the "/transaction/cosignature" HTTP response.

    :param status: Status code for HTTP response.
    :param json: JSON data for response message.
    """

    assert ((status == 200) | (status == 202))
    return models.TransactionAnnounceResponse.create_from_dto(json)


announce_cosignature = request("announce_cosignature")


# FORWARDERS
# ----------

CLIENT_CB = {
    # ACCOUNT
    'get_account_info': (
        request_get_account_info,
        process_get_account_info,
    ),
    'get_accounts_info': (
        request_get_accounts_info,
        process_get_accounts_info,
    ),
    'get_account_properties': (
        request_get_account_properties,
        process_get_account_properties,
    ),
    'get_accounts_properties': (
        request_get_accounts_properties,
        process_get_accounts_properties,
    ),
    'get_multisig_account_info': (
        request_get_multisig_account_info,
        process_get_multisig_account_info,
    ),
    'get_multisig_account_graph_info': (
        request_get_multisig_account_graph_info,
        process_get_multisig_account_graph_info,
    ),
    'get_account_transactions': (
        request_get_account_transactions,
        process_get_account_transactions,
    ),
    'get_account_incoming_transactions': (
        request_get_account_incoming_transactions,
        process_get_account_incoming_transactions,
    ),
    'get_account_outgoing_transactions': (
        request_get_account_outgoing_transactions,
        process_get_account_outgoing_transactions,
    ),
    'get_account_unconfirmed_transactions': (
        request_get_account_unconfirmed_transactions,
        process_get_account_unconfirmed_transactions,
    ),
    'get_account_partial_transactions': (
        request_get_account_partial_transactions,
        process_get_account_partial_transactions,
    ),
    # 'get_account_contracts': (
    #   request_get_account_contracts,
    #   process_get_account_contracts,
    # ),
    'get_account_names': (
        request_get_account_names,
        process_get_account_names,
    ),

    # BLOCKCHAIN
    'get_block_by_height': (
        request_get_block_by_height,
        process_get_block_by_height,
    ),
    'get_blocks_by_height_with_limit': (
        request_get_blocks_by_height_with_limit,
        process_get_blocks_by_height_with_limit,
    ),
    'get_block_transactions': (
        request_get_block_transactions,
        process_get_block_transactions,
    ),
    'get_merkle_by_hash_in_block': (
        request_get_merkle_by_hash_in_block,
        process_get_merkle_by_hash_in_block,
    ),
    'get_block_receipts': (
        request_get_block_receipts,
        process_get_block_receipts,
    ),
    'get_blockchain_height': (
        request_get_blockchain_height,
        process_get_blockchain_height,
    ),
    'get_blockchain_score': (
        request_get_blockchain_score,
        process_get_blockchain_score,
    ),
    'get_diagnostic_blocks_by_height_with_limit': (
        request_get_diagnostic_blocks_by_height_with_limit,
        process_get_diagnostic_blocks_by_height_with_limit,
    ),
    'get_diagnostic_storage': (
        request_get_diagnostic_storage,
        process_get_diagnostic_storage,
    ),
    'get_diagnostic_server': (
        request_get_diagnostic_server,
        process_get_diagnostic_server,
    ),

    # CONTRACT
    # 'get_contract': (
    #    request_get_contract,
    #    process_get_contract,
    # ),
    # 'get_contracts': (
    #    request_get_contracts,
    #    process_get_contracts,
    # ),

    # METADATA
    'get_account_metadata': (
        request_get_account_metadata,
        process_get_account_metadata,
    ),
    'get_mosaic_metadata': (
        request_get_mosaic_metadata,
        process_get_mosaic_metadata,
    ),
    'get_namespace_metadata': (
        request_get_namespace_metadata,
        process_get_namespace_metadata,
    ),
    'get_metadata': (
        request_get_metadata,
        process_get_metadata,
    ),
    'get_metadatas': (
        request_get_metadatas,
        process_get_metadatas,
    ),

    # CONFIG
    'get_config': (
        request_get_config,
        process_get_config,
    ),
    'get_upgrade': (
        request_get_upgrade,
        process_get_upgrade,
    ),

    # NODE
    'get_node_info': (
        request_get_node_info,
        process_get_node_info,
    ),
    'get_node_time': (
        request_get_node_time,
        process_get_node_time,
    ),

    # MOSAIC
    'get_mosaic': (
        request_get_mosaic,
        process_get_mosaic,
    ),
    'get_mosaics': (
        request_get_mosaics,
        process_get_mosaics,
    ),
    'get_mosaic_names': (
        request_get_mosaic_names,
        process_get_mosaic_names,
    ),
    'get_mosaic_richlist': (
        request_get_mosaic_richlist,
        process_get_mosaic_richlist,
    ),

    # NAMESPACE
    'get_namespace': (
        request_get_namespace,
        process_get_namespace,
    ),
    'get_namespaces_name': (
        request_get_namespaces_name,
        process_get_namespaces_name,
    ),
    'get_namespaces_from_account': (
        request_get_namespaces_from_account,
        process_get_namespaces_from_account,
    ),
    'get_namespaces_from_accounts': (
        request_get_namespaces_from_accounts,
        process_get_namespaces_from_accounts,
    ),
    'get_linked_mosaic_id': (
        request_get_linked_mosaic_id,
        process_get_linked_mosaic_id,
    ),
    'get_linked_address': (
        request_get_linked_address,
        process_get_linked_address,
    ),

    # NETWORK
    'get_network_type': (
        request_get_network_type,
        process_get_network_type,
    ),

    # TRANSACTOON
    'get_transaction': (
        request_get_transaction,
        process_get_transaction,
    ),
    'get_transactions': (
        request_get_transactions,
        process_get_transactions,
    ),
    'get_transaction_status': (
        request_get_transaction_status,
        process_get_transaction_status,
    ),
    'get_transaction_statuses': (
        request_get_transaction_statuses,
        process_get_transaction_statuses,
    ),
    'announce': (
        request_announce,
        process_announce,
    ),
    'announce_partial': (
        request_announce_partial,
        process_announce_partial,
    ),
    'announce_cosignature': (
        request_announce_cosignature,
        process_announce_cosignature,
    ),
}



from xpxchain import models
from tests import harness


@harness.model_test_case({
    'type': models.Mosaic,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'id': models.MosaicId(5),
        'amount': 1000,
    },
    'dto': {
        'id': [5, 0],
        'amount': [1000, 0],
    },
    'catbuffer': b'\x05\x00\x00\x00\x00\x00\x00\x00\xe8\x03\x00\x00\x00\x00\x00\x00',
})
class TestMosaic(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.MosaicId,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'id': 5,
    },
    'extras': {
        'nonce': models.MosaicNonce(b'\x00\x00\x00\x00'),
        'public_key': '7D08373CFFE4154E129E04F0827E5F3D6907587E348757B0F87D2F839BF88246',
    },
})
class TestMosaicId(harness.TestCase):

    def test_int(self):
        self.assertEqual(int(self.model), 5)

    def test_index(self):
        self.assertEqual(hex(self.model), "0x5")

    def test_format(self):
        value = self.type(13)
        self.assertEqual(f'{value:x}', 'd')
        self.assertEqual(f'{value:X}', 'D')

    def test_create_from_nonce(self):
        owner = models.PublicAccount.create_from_public_key(self.extras['public_key'], self.network_type)
        value = self.type.create_from_nonce(self.extras['nonce'], owner)
        self.assertEqual(value.id, 0x2FF7D64F483BC0A6)


@harness.model_test_case({
    'type': models.MosaicInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'meta_id': '5cc07cbc3a48065f47d6df80',
        'mosaic_id': models.MosaicId.create_from_hex('6c699a1517bea955'),
        'supply': 8999999998000000,
        'height': 1,
        'owner': models.PublicAccount.create_from_public_key('a04335f99d9ee3787528a16c7a302f80d511e9cf71d97d95c2182e0ea75a1ef9', models.NetworkType.MIJIN_TEST),
        'revision': 1,
        'properties': models.MosaicProperties(0x2, 6, 0),
    },
    'dto': {
        'meta': {
            'id': '5cc07cbc3a48065f47d6df80',
        },
        'mosaic': {
            'mosaicId': [398371157, 1818860053],
            'supply': [3403414400, 2095475],
            'height': [1, 0],
            'owner': 'a04335f99d9ee3787528a16c7a302f80d511e9cf71d97d95c2182e0ea75a1ef9',
            'revision': 1,
            'properties': [{'id': 0, 'value': [2, 0]}, {'id': 1, 'value': [6, 0]}, {'id': 2, 'value': [0, 0]}],
        },
    },
})
class TestMosaicInfo(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.MosaicName,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'mosaic_id': models.MosaicId(0xd525ad41d95fcf29),
        'names': ['xem'],
    },
    'dto': {
        'mosaicId': [0xd95fcf29, 0xd525ad41],
        'names': ['xem'],
    },
})
class TestMosaicName(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.MosaicNonce,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'nonce': b'\x12\x34\x56\x78',
    },
    'dto': 0x78563412,
    'catbuffer': b'\x12\x34\x56\x78',
    'extras': {
        'nonce': models.MosaicNonce(b'\x12\x34\x56\x78'),
        'public_key': '7D08373CFFE4154E129E04F0827E5F3D6907587E348757B0F87D2F839BF88246',
    },
})
class TestMosaicNonce(harness.TestCase):

    def test_rich_init(self):
        self.assertEqual(self.model, self.type('12345678'))
        with self.assertRaises(ValueError):
            self.type('12')
        with self.assertRaises(ValueError):
            self.type(b'\x12')
        with self.assertRaises(TypeError):
            self.type(None)

    def test_int(self):
        self.assertEqual(int(self.model), 0x78563412)

    def test_create_random(self):
        def fake_entropy(size: int):
            return b'4' * size

        self.assertEqual(self.type.create_random(fake_entropy).nonce, b'4444')
        self.type.create_random()

    def test_create_from_hex(self):
        self.assertEqual(self.model, self.type.create_from_hex('12345678'))

    def test_create_from_int(self):
        self.assertEqual(self.type.create_from_int(5).nonce, b'\x05\x00\x00\x00')
        self.assertEqual(self.type.create_from_int(325).nonce, b'E\x01\x00\x00')


@harness.model_test_case({
    'type': models.MosaicProperties,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'flags': 0x3,
        'divisibility': 1,
        'duration': 100,
    },
    'dto': [{'id': 0, 'value': [3, 0]}, {'id': 1, 'value': [1, 0]}, {'id': 2, 'value': [100, 0]}],
})
class TestMosaicProperties(harness.TestCase):

    def test_properties(self):
        self.assertEqual(self.model.supply_mutable, True)
        self.assertEqual(self.model.transferable, True)

    def test_create(self):
        self.assertEqual(self.type.create().flags, 0x2)
        self.assertEqual(self.type.create(supply_mutable=True).flags, 0x3)
        self.assertEqual(self.type.create(transferable=True).flags, 0x2)
        self.assertEqual(self.type.create(transferable=False).flags, 0x0)


@harness.enum_test_case({
    'type': models.MosaicSupplyType,
    'enums': [
        models.MosaicSupplyType.DECREASE,
        models.MosaicSupplyType.INCREASE,
    ],
    'values': [
        0,
        1,
    ],
    'descriptions': [
        'Decrease mosaic supply.',
        'Increase mosaic supply.',
    ],
    'dto': [
        0,
        1,
    ],
    'catbuffer': [
        b'\x00',
        b'\x01',
    ],
})
class TestMosaicSupplyType(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.NetworkCurrencyMosaic,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'amount': 1,
    },
    'dto': {
        'amount': [1, 0],
        'id': [3294802500, 2243684972],
    },
    'catbuffer': b'D\xb2b\xc4l\xea\xbb\x85\x01\x00\x00\x00\x00\x00\x00\x00',
})
class TestNetworkCurrencyMosaic(harness.TestCase):

    def test_class_variables(self):
        self.assertEqual(self.type.NAMESPACE_ID, models.NamespaceId(0x85BBEA6CC462B244))
        self.assertEqual(self.type.DIVISIBILITY, 6)
        self.assertEqual(self.type.INITIAL_SUPPLY, 8999999998)
        self.assertEqual(self.type.TRANSFERABLE, True)
        self.assertEqual(self.type.SUPPLY_MUTABLE, False)
        self.assertEqual(self.type.LEVY_MUTABLE, False)

    def test_create_relative(self):
        self.assertEqual(self.type.create_relative(1).amount, 1000000)

    def test_create_absolute(self):
        self.assertEqual(self.type.create_absolute(1).amount, 1)

    def test_from_invalid_dto(self):
        with self.assertRaises(ValueError):
            self.type.create_from_dto({'amount': [1, 0], 'id': [0, 0]})

    def test_from_invalid_catbuffer(self):
        with self.assertRaises(ValueError):
            catbuffer = bytes(8) + self.catbuffer[8:]
            self.type.create_from_catbuffer(catbuffer)


@harness.model_test_case({
    'type': models.NetworkHarvestMosaic,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'amount': 1,
    },
    'dto': {
        'amount': [1, 0],
        'id': [3084986652, 2484246962],
    },
    'catbuffer': b'\x1c)\xe1\xb7\xb2\x99\x12\x94\x01\x00\x00\x00\x00\x00\x00\x00',
})
class TestNetworkHarvestMosaic(harness.TestCase):

    def test_class_variables(self):
        self.assertEqual(self.type.NAMESPACE_ID, models.NamespaceId(0x941299B2B7E1291C))
        self.assertEqual(self.type.DIVISIBILITY, 3)
        self.assertEqual(self.type.INITIAL_SUPPLY, 15000000)
        self.assertEqual(self.type.TRANSFERABLE, True)
        self.assertEqual(self.type.SUPPLY_MUTABLE, True)
        self.assertEqual(self.type.LEVY_MUTABLE, False)

    def test_create_relative(self):
        self.assertEqual(self.type.create_relative(1).amount, 1000)

    def test_create_absolute(self):
        self.assertEqual(self.type.create_absolute(1).amount, 1)

    def test_from_invalid_dto(self):
        with self.assertRaises(ValueError):
            self.type.create_from_dto({'amount': [1, 0], 'id': [0, 0]})

    def test_from_invalid_catbuffer(self):
        with self.assertRaises(ValueError):
            catbuffer = bytes(8) + self.catbuffer[8:]
            self.type.create_from_catbuffer(catbuffer)



"""
    hashlib
    =======

    Cryptographic hash functions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .keccak import keccak_224, keccak_256, keccak_384, keccak_512
from .ripemd import ripemd160
from .sha2 import sha224, sha256, sha384, sha512
from .sha3 import sha3_224, sha3_256, sha3_384, sha3_512

__all__ = [
    'keccak_224',
    'keccak_256',
    'keccak_384',
    'keccak_512',
    'ripemd160',
    'sha224',
    'sha256',
    'sha384',
    'sha512',
    'sha3_224',
    'sha3_256',
    'sha3_384',
    'sha3_512',
]



from xpxchain import models
from xpxchain import util
from xpxchain.models.transaction.recipient import Recipient
from tests import harness


class TestRecipient(harness.TestCase):

    def setUp(self):
        self.network_type = models.NetworkType.MIJIN_TEST
        self.address = models.Address("SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54")
        self.namespace_id = models.NamespaceId(0x84b3552d375ffa4b)
        self.address_dto = '90fa39ec47e05600afa74308a7ea607d145e371b5f4f1447bc'
        self.namespace_dto = '914bfa5f372d55b38400000000000000000000000000000000'
        self.address_catbuffer = util.unhexlify(self.address_dto)
        self.namespace_catbuffer = util.unhexlify(self.namespace_dto)

    def test_to_dto(self):
        self.assertEqual(self.address_dto, Recipient.to_dto(self.address, self.network_type))
        self.assertEqual(self.namespace_dto, Recipient.to_dto(self.namespace_id, self.network_type))

    def test_from_dto(self):
        self.assertEqual(self.address, Recipient.create_from_dto(self.address_dto, self.network_type))
        self.assertEqual(self.namespace_id, Recipient.create_from_dto(self.namespace_dto, self.network_type))

    def test_to_catbuffer(self):
        self.assertEqual(self.address_catbuffer, Recipient.to_catbuffer(self.address, self.network_type))
        self.assertEqual(self.namespace_catbuffer, Recipient.to_catbuffer(self.namespace_id, self.network_type))

    def test_create_from_catbuffer(self):
        self.assertEqual(self.address, Recipient.create_from_catbuffer(self.address_catbuffer, self.network_type))
        self.assertEqual(self.namespace_id, Recipient.create_from_catbuffer(self.namespace_catbuffer, self.network_type))



from xpxchain import util
from tests import harness


def test_stdint(self, x: int, name: str):
    getattr(util, f'{name}_high')(x)
    getattr(util, f'{name}_low')(x)

    dto = getattr(util, f'{name}_to_dto')(x)
    self.assertEqual(getattr(util, f'{name}_from_dto')(dto), x)

    catbuffer = getattr(util, f'{name}_to_catbuffer')(x)
    self.assertEqual(getattr(util, f'{name}_from_catbuffer')(catbuffer), x)


def test_stdint_error(self, x: int, name: str):
    with self.assertRaises(ArithmeticError):
        getattr(util, f'{name}_high')(x)
    with self.assertRaises(ArithmeticError):
        getattr(util, f'{name}_low')(x)
    with self.assertRaises(ArithmeticError):
        dto = getattr(util, f'{name}_to_dto')(x)
        self.assertEqual(getattr(util, f'{name}_from_dto')(dto), x)
    with self.assertRaises(ArithmeticError):
        catbuffer = getattr(util, f'{name}_to_catbuffer')(x)
        self.assertEqual(getattr(util, f'{name}_from_catbuffer')(catbuffer), x)


class TestUnsignedInt(harness.TestCase):

    @harness.randomize
    def test_u8(self, x: harness.U8):
        test_stdint(self, x, 'u8')

    @harness.randomize(x={'min_value': -1 << 8, 'max_value': -1})
    def test_u8_underflow(self, x: int):
        test_stdint_error(self, x, 'u8')

    @harness.randomize(x={'min_value': 1 << 8, 'max_value': 1 << 9})
    def test_u8_overflow(self, x: int):
        test_stdint_error(self, x, 'u8')

    @harness.randomize
    def test_u16(self, x: harness.U16):
        test_stdint(self, x, 'u16')

    @harness.randomize(x={'min_value': -1 << 16, 'max_value': -1})
    def test_u16_underflow(self, x: int):
        test_stdint_error(self, x, 'u16')

    @harness.randomize(x={'min_value': 1 << 16, 'max_value': 1 << 17})
    def test_u16_overflow(self, x: int):
        test_stdint_error(self, x, 'u16')

    @harness.randomize
    def test_u32(self, x: harness.U32):
        test_stdint(self, x, 'u32')

    @harness.randomize(x={'min_value': -1 << 32, 'max_value': -1})
    def test_u32_underflow(self, x: int):
        test_stdint_error(self, x, 'u32')

    @harness.randomize(x={'min_value': 1 << 32, 'max_value': 1 << 33})
    def test_u32_overflow(self, x: int):
        test_stdint_error(self, x, 'u32')

    @harness.randomize
    def test_u64(self, x: harness.U64):
        test_stdint(self, x, 'u64')

    @harness.randomize(x={'min_value': -1 << 64, 'max_value': -1})
    def test_u64_underflow(self, x: int):
        test_stdint_error(self, x, 'u64')

    @harness.randomize(x={'min_value': 1 << 64, 'max_value': 1 << 65})
    def test_u64_overflow(self, x: int):
        test_stdint_error(self, x, 'u64')

    @harness.randomize
    def test_u128(self, x: harness.U128):
        test_stdint(self, x, 'u128')

    @harness.randomize(x={'min_value': -1 << 128, 'max_value': -1})
    def test_u128_underflow(self, x: int):
        test_stdint_error(self, x, 'u128')

    @harness.randomize(x={'min_value': 1 << 128, 'max_value': 1 << 129})
    def test_u128_overflow(self, x: int):
        test_stdint_error(self, x, 'u128')



import binascii
import warnings

from xpxchain.util.signature.ed25519 import sha3 as ed25519
from tests import harness

# DUMMY DATA
# Never use this private key for real data.
SIGNING_KEY = '97131746d864f4c9001b1b86044d765ba08d7fddc7a0fb3abbc8d111aa26cdca1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955'
VERIFYING_KEY = '1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955'
MESSAGE = '68656c6c6f20776f726c64'
SIGNATURE = '80b2168f89f1197b3c00cb8555ada77a5866e5d14ebdc907467fe4ad2da204348333adcdf3a267b395e3a6b4a50ce9021a5a017885a5882f33a6d1c8f64a2d0d'


class TestEd25519Sha3(harness.TestCase):

    def setUp(self):
        self.signing_key = ed25519.SigningKey(binascii.unhexlify(SIGNING_KEY))
        self.verifying_key = ed25519.VerifyingKey(binascii.unhexlify(VERIFYING_KEY))
        self.message = binascii.unhexlify(MESSAGE)
        self.signature = binascii.unhexlify(SIGNATURE)

    def test_create_keypair(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # dummy entropy function so we can use existing seed
            entropy = lambda x: self.signing_key.to_seed()[:x]
            signing_key, verifying_key = ed25519.create_keypair(entropy)
            self.assertEqual(signing_key, self.signing_key)
            self.assertEqual(verifying_key, self.verifying_key)

    def test_from_seed(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            seed = self.signing_key.to_seed()
            signing_key = ed25519.SigningKey(seed)
            self.assertEqual(signing_key, self.signing_key)
            self.assertEqual(signing_key.get_verifying_key(), self.verifying_key)

    def test_sign(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            signature = self.signing_key.sign(self.message)
            self.assertEqual(signature, self.signature)

    def test_verify(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            self.verifying_key.verify(self.signature, self.message)

            with self.assertRaises(ed25519.BadSignatureError):
                signature = self.signature[:-1] + b'0'
                self.verifying_key.verify(signature, self.message)

            with self.assertRaises((AssertionError, ValueError)):
                self.verifying_key.verify(self.signature + b'0', self.message)
            with self.assertRaises((AssertionError, ValueError)):
                self.verifying_key.verify(self.signature[:-1], self.message)



"""
    plain_message
    =============

    Plain message, sent to the network as hex-encoded bytes.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .message import Message
from .message_type import MessageType
from ... import util

__all__ = ['PlainMessage']


@util.inherit_doc
class PlainMessage(Message):
    """
    Defines a plain message object. Sent to the network as hex.

    :param payload: Message data, as bytes or hex.
    """

    __slots__ = ()

    def __init__(self, payload: typing.AnyStr) -> None:
        payload = util.decode_hex(payload, with_prefix=True)
        super().__init__(MessageType.PLAIN, payload)

    @classmethod
    def create(cls, payload: typing.AnyStr):
        return cls(payload)  # type: ignore


EMPTY_MESSAGE = PlainMessage(b'')



"""
    network_currency_mosaic
    =======================

    Description of a per-network harvest mosaic.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .mosaic import Mosaic
from ..blockchain.network_type import OptionalNetworkType
from ..namespace.namespace_id import NamespaceId
from ... import util

__all__ = ['NetworkHarvestMosaic']


class NetworkHarvestMosaic(Mosaic):
    """
    Per-network harvest mosaic.

    :param amount: Mosaic quantity in the smallest unit possible.
    :cvar NAMESPACE_ID: Namespace identifier for asset.
    :cvar DIVISIBILITY: Decimal place mosaic can be divided into.
    :cvar INITIAL_SUPPLY: Initial supply of asset.
    :cvar TRANSFERABLE: Allow transfer of funds from accounts other than creator.
    :cvar SUPPLY_MUTABLE: Mosaic allows supply change later.
    :cvar LEVY_MUTABLE: If levy is mutable.
    """

    __slots__ = ()
    NAMESPACE_ID: typing.ClassVar[NamespaceId] = NamespaceId("cat.harvest")
    DIVISIBILITY: typing.ClassVar[int] = 3
    INITIAL_SUPPLY: typing.ClassVar[int] = 15000000
    TRANSFERABLE: typing.ClassVar[bool] = True
    SUPPLY_MUTABLE: typing.ClassVar[bool] = True
    LEVY_MUTABLE: typing.ClassVar[bool] = False

    def __init__(self, amount: int) -> None:
        super().__init__(self.NAMESPACE_ID, amount)

    @classmethod
    def create_relative(cls, amount: int):
        """
        Create `NetworkHarvestMosaic` using relative (divisibility) units.

        :param amount: Mosaic quantity in relative units.

        Example:
            .. code-block:: python

                >>> NetworkHarvestMosaic.create_relative(1).amount
                1000
        """
        return cls(amount * 10 ** cls.DIVISIBILITY)

    @classmethod
    def create_absolute(cls, amount: int):
        """
        Create `NetworkHarvestMosaic` using absolute (smallest) units.

        :param amount: Mosaic quantity in absolute units.


        Example:
            .. code-block:: python

                >>> NetworkHarvestMosaic.create_relative(1).amount
                1
        """
        return cls(amount)

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        id = NamespaceId(util.u64_from_dto(data['id']))
        amount = util.u64_from_dto(data['amount'])
        if int(id) != int(cls.NAMESPACE_ID):
            raise ValueError('Network harvest mosaic ID does not match.')
        return cls(amount)

    @classmethod
    def create_from_catbuffer(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        id = NamespaceId(util.u64_from_catbuffer(data[:8]))
        amount = util.u64_from_catbuffer(data[8:16])
        if int(id) != int(cls.NAMESPACE_ID):
            raise ValueError('Network harvest mosaic ID does not match.')
        return cls(amount)



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.ConfigHTTP, client.AsyncConfigHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_config',
            'response': responses.CONFIG["Ok"],
            'params': [100],
            'method': 'get_config',
            'validation': [
                lambda x: (isinstance(x, models.CatapultConfig), True),
                lambda x: (x.height, 1),
                lambda x: (x.network_config, 'config'),
                lambda x: (x.supported_entity_versions, 'versions'),
            ],
        },
        {
            'name': 'test_get_upgrade',
            'response': responses.UPGRADE["Ok"],
            'params': [100],
            'method': 'get_upgrade',
            'validation': [
                lambda x: (isinstance(x, models.CatapultUpgrade), True),
                lambda x: (isinstance(x.blockchain_upgrade, models.Upgrade), True),
                lambda x: (x.blockchain_upgrade.height, 1),
                lambda x: (x.blockchain_upgrade.blockchain_version, 17180000256),
            ],
        },
    ],
})
class TestConfigHTTP(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses
from xpxchain import util


@harness.mocked_http_test_case({
    'clients': (client.TransactionHTTP, client.AsyncTransactionHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_transaction',
            'response': responses.TRANSACTION["Ok"],
            'params': ['47490969DB1960AD8565E67700C47FE41BBE07C6F490A66D3001AC46B6684600'],
            'method': 'get_transaction',
            'validation': [
                lambda x: (isinstance(x, models.Transaction), True),
                lambda x: (isinstance(x, models.TransferTransaction), True),
                lambda x: (isinstance(x.type, models.TransactionType), True),
                lambda x: (x.type, models.TransactionType.TRANSFER),
                lambda x: (isinstance(x.network_type, models.NetworkType), True),
                lambda x: (x.network_type, models.NetworkType.MIJIN_TEST),
                lambda x: (x.version, models.TransactionVersion.TRANSFER),
                lambda x: (isinstance(x.deadline, models.Deadline), True),
                lambda x: (x.deadline, models.Deadline.create_from_timestamp(util.u64_from_dto([2208860662, 28]))),
                lambda x: (x.max_fee, 41750),
                lambda x: (x.signature, '3A6D4A8EE69C509F7D020C5DBCD3B7E9610B0571F2C7F14229D900A85D51E41EB52BCBEFFD1D95CB40C996540657DBA33090E27A3AA5E0D71F727A055B6D5403'),
                lambda x: (isinstance(x.signer, models.PublicAccount), True),
                lambda x: (x.signer.public_key, '0EB448D07C7CCB312989AC27AA052738FF589E2F83973F909B506B450DC5C4E2'),
                lambda x: (isinstance(x.transaction_info, models.TransactionInfo), True),
                lambda x: (x.transaction_info.height, 158158),
                lambda x: (x.transaction_info.hash, '47490969DB1960AD8565E67700C47FE41BBE07C6F490A66D3001AC46B6684600'),
                lambda x: (x.transaction_info.id, '5E4A463F6466D3000169F159'),
                lambda x: (x.transaction_info.index, 0),
                lambda x: (x.transaction_info.merkle_component_hash, '47490969DB1960AD8565E67700C47FE41BBE07C6F490A66D3001AC46B6684600'),
                lambda x: (isinstance(x.recipient, models.Address), True),
                lambda x: (x.recipient.hex, '90C9102514E6A3893E53A5E06365E1074462E3BB1AF4F36529'),
                lambda x: (len(x.mosaics), 1),
                lambda x: (isinstance(x.mosaics[0], models.Mosaic), True),
                lambda x: (isinstance(x.mosaics[0].id, models.MosaicId), True),
                lambda x: (int(x.mosaics[0].id), 992621222383397347),
                lambda x: (x.mosaics[0].amount, 10000000),
                lambda x: (isinstance(x.message, models.Message), True),
                lambda x: (isinstance(x.message.type, models.MessageType), True),
                lambda x: (x.message.type, models.MessageType.PLAIN),
                lambda x: (x.message.payload, b''),
            ],
        },
        {
            'name': 'test_get_transactions',
            'response': responses.TRANSACTIONS["Ok"],
            'params': [['47490969DB1960AD8565E67700C47FE41BBE07C6F490A66D3001AC46B6684600']],
            'method': 'get_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (isinstance(x[0], models.Transaction), True),
                lambda x: (isinstance(x[0], models.TransferTransaction), True),
            ],
        },
        {
            'name': 'test_get_transaction_status',
            'response': responses.TRANSACTION_STATUS["Ok"],
            'params': ['135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9'],
            'method': 'get_transaction_status',
            'validation': [
                lambda x: (isinstance(x, models.TransactionStatus), True),
                lambda x: (isinstance(x.group, models.TransactionStatusGroup), True),
                lambda x: (x.group, models.TransactionStatusGroup.CONFIRMED),
                lambda x: (x.status, 'Success'),
                lambda x: (x.hash, '135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9'),
                lambda x: (isinstance(x.deadline, models.Deadline), True),
                lambda x: (x.deadline, models.Deadline.create_from_timestamp(util.u64_from_dto([2208867673, 28]))),
                lambda x: (x.height, 158159),
            ],
        },
        {
            'name': 'test_get_transaction_statuses',
            'response': responses.TRANSACTION_STATUSES["Ok"],
            'params': [['135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9']],
            'method': 'get_transaction_statuses',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (isinstance(x[0], models.TransactionStatus), True),
            ],
        },
        {
            'name': 'test_announce',
            'response': responses.ANNOUNCE["Ok"],
            'params': [
                models.SignedTransaction(
                    payload='a70000004b94...',
                    hash='135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9',
                    signer='0EB448D07C7CCB312989AC27AA052738FF589E2F83973F909B506B450DC5C4E2',
                    type=models.TransactionType.TRANSFER,
                    network_type=models.NetworkType.MIJIN_TEST,
                )
            ],
            'method': 'announce',
            'validation': [
                lambda x: (isinstance(x, models.TransactionAnnounceResponse), True),
                lambda x: (x.message, 'packet 9 was pushed to the network via /transaction'),
            ],
        },
        {
            'name': 'test_announce_partial',
            'response': responses.ANNOUNCE_PARTIAL["Ok"],
            'params': [
                models.SignedTransaction(
                    payload='a70000004b94...',
                    hash='135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9',
                    signer='0EB448D07C7CCB312989AC27AA052738FF589E2F83973F909B506B450DC5C4E2',
                    type=models.TransactionType.TRANSFER,
                    network_type=models.NetworkType.MIJIN_TEST,
                )
            ],
            'method': 'announce_partial',
            'validation': [
                lambda x: (isinstance(x, models.TransactionAnnounceResponse), True),
                lambda x: (x.message, 'packet 500 was pushed to the network via /transaction/partial'),
            ],
        },
        {
            'name': 'test_announce_cosignature',
            'response': responses.ANNOUNCE_COSIGNATURE["Ok"],
            'params': [
                models.CosignatureSignedTransaction(
                    parent_hash='135B65FF4E9E90E2F3283D75E2AD3D45EC46729013786841E32904D5134FD2B9',
                    signature='3A6D4A8EE69C509F7D020C5DBCD3B7E9610B0571F2C7F14229D900A85D51E41EB52BCBEFFD1D95CB40C996540657DBA33090E27A3AA5E0D71F727A055B6D5403',
                    signer='0EB448D07C7CCB312989AC27AA052738FF589E2F83973F909B506B450DC5C4E2',
                )
            ],
            'method': 'announce_cosignature',
            'validation': [
                lambda x: (isinstance(x, models.TransactionAnnounceResponse), True),
                lambda x: (x.message, 'packet 501 was pushed to the network via /transaction/cosignature'),
            ],
        },
    ],
})
class TestTransactionHTTP(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.NodeHTTP, client.AsyncNodeHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_node_info',
            'response': responses.INFO["Ok"],
            'params': [],
            'method': 'get_node_info',
            'validation': [
                lambda x: (isinstance(x, models.NodeInfo), True),
                lambda x: (x.public_key, '460458B98E2BAA36A8E95DE9B320379E89898885B71CF0174E02F1324FAFFAC1'),
                lambda x: (x.port, 7900),
                lambda x: (x.network_identifier, models.NetworkType.MIJIN_TEST),
                lambda x: (x.version, 0),
                lambda x: (x.roles, 2),
                lambda x: (x.friendly_name, 'api-node-0'),
            ],
        },
        {
            'name': 'test_get_node_time',
            'response': responses.TIME["Ok"],
            'params': [],
            'method': 'get_node_time',
            'validation': [
                lambda x: (isinstance(x, models.NodeTime), True),
                lambda x: (x.send_timestamp, 122457123440),
                lambda x: (x.receive_timestamp, 122457123440),
            ],
        },
    ],
})
class TestConfigHTTP(harness.TestCase):
    pass



from xpxchain import models
from xpxchain import client
import os
import requests

gen_hash = None
mosaic_id = None
divisibility = None
tester = None
nemesis = None

if (not tester):

    ENDPOINT = os.environ.get('ENDPOINT', '//localhost:3000')
    FAUCET = os.environ.get('FAUCET', '')
    PRIVATE_KEY = os.environ.get('PRIVATE_KEY', '')

    print()
    print(f"Endpoint: '{ENDPOINT}'")
    if (FAUCET):
        print(f"  Faucet: '{FAUCET}'")

    if (not ENDPOINT):
        raise Exception('No endpoint provided.')

    if ((not FAUCET) and (not PRIVATE_KEY)):
        raise Exception('Neither of faucet or private key were provided.')

    # Get network type
    with client.NodeHTTP(ENDPOINT) as http:
        node_info = http.get_node_info()
        network_type = node_info.network_identifier

    # Get generation hash and main mosaic
    with client.BlockchainHTTP(ENDPOINT) as http:
        block = http.get_block_by_height(1)
        gen_hash = block.generation_hash
        nemesis = block.signer

        receipts = http.get_block_receipts(2)
        for statement in receipts.transaction_statements:
            for receipt in statement.receipts:
                if (receipt.type == models.ReceiptType.VALIDATE_FEE):
                    mosaic_id = receipt.mosaic.id
                    break

    # Checking the divisibility
    with client.MosaicHTTP(ENDPOINT) as http:
        mosaic_info = http.get_mosaic(mosaic_id)
        divisibility = 10 ** mosaic_info.properties.divisibility

    # Prepare the testing account
    if (PRIVATE_KEY):
        account = models.Account.create_from_private_key(PRIVATE_KEY, network_type)
    else:
        account = models.Account.generate_new_account(network_type, entropy=lambda x: os.urandom(32))
        reply = requests.get(f"{FAUCET}{account.address.address}").json()
        if (reply != 'XPX sent!'):
            raise Exception(f"Error getting funds from faucet: {reply}")

    print()
    print(f"Testing account")
    print(f"---------------")
    print(f"    Address: {account.address.address}")
    print(f"Private key: {account.private_key}")
    print(f" Public key: {account.public_key}")
    print()

    tester = account



import asyncio
import aiohttp
import requests

from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


class TestAsyncLoop(harness.TestCase):

    def test_loop(self):
        async def inner(loop):
            async with client.AsyncHTTP(responses.ENDPOINT, loop=loop) as http:
                # Set the network type
                with aiohttp.default_response(200, **responses.NETWORK_TYPE["MIJIN_TEST"]):
                    await http.network_type

                with aiohttp.default_response(200, **responses.BLOCK_INFO["Ok"]):
                    block_info = await http.blockchain.get_block_by_height(1)
                    self.assertEqual(block_info.total_fee, 0)

        loop = asyncio.new_event_loop()
        loop.run_until_complete(inner(loop))
        loop.close()


class TestHTTP(harness.TestCase):

    @harness.async_test(
        sync_data=(client.NetworkHTTP, requests, client.HTTPError),
        async_data=(client.AsyncNetworkHTTP, aiohttp, client.AsyncHTTPError)
    )
    async def test_exceptions(self, data, await_cb, with_cb):
        async with with_cb(data[0](responses.ENDPOINT)) as http:
            with self.assertRaises(data[2]):
                with data[1].default_exception(data[2]):
                    await await_cb(http.get_network_type())
            with self.assertRaises(ConnectionRefusedError):
                with data[1].default_exception(ConnectionRefusedError):
                    await await_cb(http.get_network_type())

    @harness.async_test(
        sync_data=client.HTTP,
        async_data=client.AsyncHTTP
    )
    async def test_create_from_http(self, data, await_cb, with_cb):
        async with with_cb(data(responses.ENDPOINT)) as http:
            copy = data.create_from_http(http)
            self.assertTrue(http.raw is copy.raw)

    @harness.async_test(
        sync_data=(client.HTTP, requests),
        async_data=(client.AsyncHTTP, aiohttp)
    )
    async def test_network_type(self, data, await_cb, with_cb):
        async with with_cb(data[0](responses.ENDPOINT)) as http:
            with data[1].default_response(200, **responses.NETWORK_TYPE["MIJIN_TEST"]):
                self.assertEqual(await await_cb(http.network_type), models.NetworkType.MIJIN_TEST)



"""
    bit
    ===

    Utilities to modify individual bits from bytes.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations


def set(byte: int, index: int) -> int:
    """Set bit at index to 1."""

    assert 0 <= byte <= 255
    assert 0 <= index <= 7

    return byte | (1 << index)


def clear(byte: int, index: int) -> int:
    """Set bit at index to 0."""

    assert 0 <= byte <= 255
    assert 0 <= index <= 7

    # Python guarantees the ~ operator will return the 2s complement
    # signed integer with the same bit pattern, AKA, ~128 is -129, not 127.
    # However, the & operator with the resulting value will use the sign bit,
    # so `255 & 127 == 255 & -129`, for example.
    return byte & ~(1 << index)


def assign(byte: int, index: int, value: bool) -> int:
    """Assign bit at index depending on value."""

    if value:
        return set(byte, index)
    else:
        return clear(byte, index)


def toggle(byte: int, index: int) -> int:
    """Toggle bit at index."""

    assert 0 <= byte <= 255
    assert 0 <= index <= 7

    return byte ^ (1 << index)


def get(byte: int, index: int) -> int:
    """Get bit at index."""

    assert 0 <= byte <= 255
    assert 0 <= index <= 7

    return (byte >> index) & 1



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses
from xpxchain import util


@harness.mocked_http_test_case({
    'clients': (client.AccountHTTP, client.AsyncAccountHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_account_info',
            'response': responses.ACCOUNT_INFO["Ok"],
            'params': [models.Address('SCBO3CAFOVAOGYBAHQKPUOGLAYWFLNJUFFCH3RYY')],
            'method': 'get_account_info',
            'validation': [
                lambda x: (x.public_key, '7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808'),
                lambda x: (x.mosaics, []),
            ],
        },
        {
            'name': 'test_get_accounts_info',
            'response': responses.ACCOUNTS_INFO["Ok"],
            'params': [[models.Address('SCBO3CAFOVAOGYBAHQKPUOGLAYWFLNJUFFCH3RYY')]],
            'method': 'get_accounts_info',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].public_key, '7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808'),
                lambda x: (x[0].mosaics, []),
            ],
        },
        {
            'name': 'test_get_account_properties',
            'response': responses.ACCOUNT_PROPERTIES["Ok"],
            'params': [models.Address.create_from_encoded('906DE63B4FDC66936F823B1CA8B08992C6FBCF0D0AEA2310AF')],
            'method': 'get_account_properties',
            'validation': [
                lambda x: (isinstance(x, models.AccountProperties), True),
                lambda x: (isinstance(x.address, models.Address), True),
                lambda x: (x.address.hex, '906DE63B4FDC66936F823B1CA8B08992C6FBCF0D0AEA2310AF'),
                lambda x: (len(x.properties), 3),
                lambda x: (isinstance(x.properties[0], models.AccountProperty), True),
                lambda x: (x.properties[0].property_type, models.PropertyType.ALLOW_ADDRESS),
                lambda x: (len(x.properties[0].values), 1),
                lambda x: (isinstance(x.properties[0].values[0], models.Address), True),
                lambda x: (x.properties[0].values[0].hex, '905F14582E43F82A7C8660D55F441C38AB3F0B630B30B21066'),
                lambda x: (x.properties[1].property_type, models.PropertyType.ALLOW_MOSAIC),
                lambda x: (len(x.properties[1].values), 1),
                lambda x: (isinstance(x.properties[1].values[0], models.MosaicId), True),
                lambda x: (int(x.properties[1].values[0]), util.u64_from_dto([481110499, 231112638])),
                lambda x: (x.properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION),
                lambda x: (len(x.properties[2].values), 0),
            ],
        },
        {
            'name': 'test_get_accounts_properties',
            'response': responses.ACCOUNTS_PROPERTIES["Ok"],
            'params': [[
                models.Address.create_from_encoded('906DE63B4FDC66936F823B1CA8B08992C6FBCF0D0AEA2310AF'),
                models.Address.create_from_encoded('907833A9D4D5EC393ADB9FE69AC7558AA397EC33094174D6C3'),
            ]],
            'method': 'get_accounts_properties',
            'validation': [
                lambda x: (len(x), 2),
                lambda x: (isinstance(x[0], models.AccountProperties), True),
                lambda x: (isinstance(x[0].address, models.Address), True),
                lambda x: (x[0].address.hex, '906DE63B4FDC66936F823B1CA8B08992C6FBCF0D0AEA2310AF'),
                lambda x: (x[1].address.hex, '907833A9D4D5EC393ADB9FE69AC7558AA397EC33094174D6C3'),
                lambda x: (len(x[0].properties), 3),
                lambda x: (isinstance(x[0].properties[0], models.AccountProperty), True),
                lambda x: (x[0].properties[0].property_type, models.PropertyType.ALLOW_ADDRESS),
                lambda x: (len(x[0].properties[0].values), 1),
                lambda x: (isinstance(x[0].properties[0].values[0], models.Address), True),
                lambda x: (x[0].properties[0].values[0].hex, '905F14582E43F82A7C8660D55F441C38AB3F0B630B30B21066'),
                lambda x: (x[0].properties[1].property_type, models.PropertyType.ALLOW_MOSAIC),
                lambda x: (len(x[0].properties[1].values), 1),
                lambda x: (isinstance(x[0].properties[1].values[0], models.MosaicId), True),
                lambda x: (int(x[0].properties[1].values[0]), util.u64_from_dto([481110499, 231112638])),
                lambda x: (x[0].properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION),
                lambda x: (len(x[0].properties[2].values), 0),
                lambda x: (len(x[1].properties), 3),
                lambda x: (isinstance(x[1].properties[0], models.AccountProperty), True),
                lambda x: (x[1].properties[0].property_type, models.PropertyType.BLOCK_ADDRESS),
                lambda x: (len(x[1].properties[0].values), 1),
                lambda x: (isinstance(x[1].properties[0].values[0], models.Address), True),
                lambda x: (x[1].properties[0].values[0].hex, '905F14582E43F82A7C8660D55F441C38AB3F0B630B30B21066'),
                lambda x: (x[1].properties[1].property_type, models.PropertyType.BLOCK_MOSAIC),
                lambda x: (len(x[1].properties[1].values), 1),
                lambda x: (isinstance(x[1].properties[1].values[0], models.MosaicId), True),
                lambda x: (int(x[1].properties[1].values[0]), util.u64_from_dto([481110499, 231112638])),
                lambda x: (x[1].properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION),
                lambda x: (len(x[1].properties[2].values), 1),
                lambda x: (isinstance(x[1].properties[2].values[0], models.TransactionType), True),
                lambda x: (x[1].properties[2].values[0], models.TransactionType.AGGREGATE_COMPLETE),
            ],
        },
        {
            'name': 'test_get_account_names',
            'response': responses.ACCOUNT_NAMES["Ok"],
            'params': [[models.Address.create_from_encoded('904E828FB27B84589C2FCFFC3B09CBC9E56C6612E4E133A558')]],
            'method': 'get_account_names',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].address.hex, '904E828FB27B84589C2FCFFC3B09CBC9E56C6612E4E133A558'),
                lambda x: (len(x[0].names), 1),
                lambda x: (x[0].names[0], 'foo.bar'),
            ],
        },
        {
            'name': 'test_get_multisig_account_info',
            'response': responses.MULTISIG_INFO["Ok"],
            'params': [models.Address('SA2AK4GEVSAX4NNG7KH4EGIXWNHMGEMWU7JT4RU5')],
            'method': 'get_multisig_account_info',
            'validation': [
                lambda x: (x.account.public_key, '9B1D7F4F4C2A0471EDBC78F95DE54B6E432887FC0C6E4CECA800089DAE2A4044'),
                lambda x: (x.min_approval, 2),
                lambda x: (x.min_removal, 2),
                lambda x: (len(x.cosignatories), 3),
                lambda x: (x.cosignatories[0].public_key, '6CB4CAAACCA7081C9E1471DF8F6512ABC99FEB86EFEE7862ED7259397C5FDBDD'),
                lambda x: (x.cosignatories[1].public_key, 'C5C55181284607954E56CD46DE85F4F3EF4CC713CC2B95000FA741998558D268'),
                lambda x: (x.cosignatories[2].public_key, 'CAE725538EBEBF7778257A442B4A48E116636580ED630AD5CB8D668DFF52A1A7'),
                lambda x: (len(x.multisig_accounts), 0),
            ],
        },
        {
            'name': 'test_get_multisig_account_graph_info',
            'response': responses.MULTISIG_GRAPH_INFO["Ok"],
            'params': [models.Address('SA2AK4GEVSAX4NNG7KH4EGIXWNHMGEMWU7JT4RU5')],
            'method': 'get_multisig_account_graph_info',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (len(x[0]), 1),
                lambda x: (x[0][0].account.public_key, '9b1d7f4f4c2a0471edbc78f95de54b6e432887fc0c6e4ceca800089dae2a4044'),
                lambda x: (x[0][0].min_approval, 2),
                lambda x: (x[0][0].min_removal, 2),
                lambda x: (len(x[0][0].cosignatories), 3),
                lambda x: (x[0][0].cosignatories[0].public_key, '6cb4caaacca7081c9e1471df8f6512abc99feb86efee7862ed7259397c5fdbdd'),
                lambda x: (x[0][0].cosignatories[1].public_key, 'c5c55181284607954e56cd46de85f4f3ef4cc713cc2b95000fa741998558d268'),
                lambda x: (x[0][0].cosignatories[2].public_key, 'cae725538ebebf7778257a442b4a48e116636580ed630ad5cb8d668dff52a1a7'),
                lambda x: (len(x[0][0].multisig_accounts), 0),
            ],
        },
        {
            'name': 'test_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].transaction_info.height, 1),
                lambda x: (x[0].transaction_info.hash, "0E814F8D301E90144F1F25384C05AC26EF9D45FACAD216590B876FBB09877E02"),
                lambda x: (x[0].transaction_info.merkle_component_hash, "0E814F8D301E90144F1F25384C05AC26EF9D45FACAD216590B876FBB09877E02"),
                lambda x: (x[0].signature, "C214F728EA91BB9947F843CB6FF59EB8D231DB652796D25D055E4D1D8E65BF67F8800B268257E9478418E1AA0D7C8DB451E82176437D21F11EB777384583230B"),
                lambda x: (x[0].signer.public_key, "A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9"),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
                lambda x: (x[0].max_fee, 0),
                lambda x: (x[0].namespace_type, models.NamespaceType.ROOT_NAMESPACE),
                lambda x: (x[0].duration, 0),
                lambda x: (x[0].namespace_id, models.NamespaceId.create_from_hex('b1497f5fba651b4f')),
                lambda x: (x[0].namespace_name, 'cat'),
            ],
        },
        {
            'name': 'test_incoming_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'incoming_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
            ],
        },
        {
            'name': 'test_outgoing_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'outgoing_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
            ],
        },
        {
            'name': 'test_outgoing_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'outgoing_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
            ],
        },
        {
            'name': 'test_unconfirmed_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'unconfirmed_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
            ],
        },
        {
            'name': 'test_aggregate_bonded_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [models.PublicAccount.create_from_public_key('7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808', models.NetworkType.MIJIN_TEST)],
            'method': 'aggregate_bonded_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
            ],
        },
    ],
})
class TestAccountHTTP(harness.TestCase):
    pass



"""
    inner_transaction
    =================

    Transaction with an embedded signer for aggregate transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import bidict
import typing

from .base import TransactionBase, TypeMap
from .format import CatbufferFormat, DTOFormat
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = ['InnerTransaction']

# We need 3 different transaction base types for `create_from_transaction`.
T1 = typing.TypeVar('T1', bound='TransactionBase')
T2 = typing.TypeVar('T2', bound='TransactionBase')
T3 = typing.TypeVar('T3', bound='TransactionBase')


@util.inherit_doc
class InnerTransaction(TransactionBase):
    """Abstract, embedded transaction base class."""

    __slots__ = ()
    # Overridable classvars.
    TYPE_MAP: typing.ClassVar[TypeMap] = bidict.bidict()
    CATBUFFER: typing.ClassVar[CatbufferFormat] = CatbufferFormat(
        # Layout
        #   uint32_t size
        #   uint8_t[32] signer
        #   uint8_t version
        #   uint8_t network_type
        #   uint16_t type
        slices={
            'size': slice(0, 4),
            'signer': slice(4, 36),
            'version': slice(36, 39),
            'network_type': slice(39, 40),
            'type': slice(40, 42),
        },
        size_shared=42,
    )
    DTO: typing.ClassVar[DTOFormat] = DTOFormat(
        names={
            'signer': 'signer',
            'version': 'version',
            'network_type': 'version',
            'type': 'type',
            'transaction_info': 'meta'
        },
    )

    # AGGREGATE

    @classmethod
    def create_from_transaction(
        cls: typing.Type[T1],
        transaction: T2,
        signer: PublicAccount,
    ) -> T3:
        """
        Generate embedded transaction from non-embedded transaction.

        :param transaction: Non-embedded transaction.
        :param signer: Account of transaction signer.
        """

        data = transaction.asdict(recurse=False)
        data['signer'] = signer
        aggregate_cls: typing.Type[T3] = cls.TYPE_MAP[data.pop('type')]
        return aggregate_cls(**data)

    # CATBUFFER

    def to_catbuffer_shared(
        self,
        network_type: NetworkType,
    ) -> bytes:
        # Shared data and callbacks.
        data = bytearray(self.catbuffer_size_shared())
        cb = lambda k, v: self.CATBUFFER.save(k, data, v, network_type)
        cb_get = lambda k: cb(k, getattr(self, k))

        # Save shared data.
        cb('size', self.catbuffer_size())
        cb_get('signer')
        cb_get('version')
        cb_get('network_type')
        cb_get('type')

        return bytes(data)

    def load_catbuffer_shared(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        # Shared data and callbacks.
        cb = lambda k: self.CATBUFFER.load(k, data, network_type)
        cb_set = lambda k: self._set(k, cb(k))

        # Load shared data.
        self._set('signature', None)
        cb_set('signer')
        cb_set('version')
        self._set('network_type', network_type)
        cb_set('type')
        self._set('max_fee', None)
        self._set('deadline', None)
        self._set('transaction_info', None)

        return data[self.catbuffer_size_shared():]

    # DTO

    @classmethod
    def validate_dto_shared(cls, data: dict) -> bool:
        required_keys = {
            'version',
            'type',
        }
        return cls.validate_dto_required(data, required_keys)

    def to_dto_shared(
        self,
        network_type: NetworkType,
    ) -> dict:
        # Shared data and callbacks.
        data: dict = {}
        cb = lambda k, v: self.DTO.save(k, data, v, network_type)
        cb_get = lambda k: cb(k, getattr(self, k))

        # Save shared data.
        cb_get('signer')
        cb_get('version')
        cb_get('network_type')
        cb_get('type')
        cb_get('transaction_info')

        return data

    def load_dto_shared(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        # Shared data and callbacks.
        cb = lambda k: self.DTO.load(k, data, network_type)
        cb_set = lambda k: self._set(k, cb(k))

        # Load shared data.
        self._set('signature', None)
        cb_set('signer')
        cb_set('version')
        self._set('network_type', network_type)
        cb_set('type')
        self._set('max_fee', None)
        self._set('deadline', None)
        self._set('transaction_info', None)


InnerTransactionList = typing.Sequence[InnerTransaction]




from __future__ import annotations

from ..blockchain.network_type import NetworkType, OptionalNetworkType
from ... import util

__all__ = ['Field']


@util.inherit_doc
@util.dataclass(frozen=True)
class Field(util.DTO):
    """

    :param key:  Key of a field.
    :param value:  Value of a field.

    DTO Format:
        .. code-block:: yaml

            FieldDTO:
                key: string
                value: string
    """

    key: str
    value: str

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {
            'key',
            'value'
        }

        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
        )

    def catbuffer_size_specific(self) -> int:
        key_size = util.U8_BYTES * len(self.key)
        value_size = util.U8_BYTES * len(self.value)

        return key_size + value_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        key = self.key.encode('utf8')
        value = self.value.encode('utf8')

        return key + value

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            key=data['key'],
            value=data['value'],
        )



"""
    abc
    ===

    Abstract base classes for HTTP and websocket clients.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import json
import typing

from . import client
from . import nis
from .. import models
from .. import util


T = typing.TypeVar('T')
MessageType = typing.Union[
    models.BlockInfo,
    models.CosignatureSignedTransaction,
    models.Transaction,
    models.TransactionStatusError,
    str,
]

# HTTP
# ----


class HTTPSharedBase(util.Object):
    """Shared, abstract base class for sync and async HTTP clients."""

    _endpoint: str
    _client: client.ClientSharedBase
    _index: int
    _network_type: typing.Optional[models.NetworkType] = None

    @property
    def index(self) -> int:
        """Get index for NIS callbacks."""
        return self._index

    @property
    def raw(self) -> client.ClientSharedBase:
        """Get client wrapper for HTTP client."""
        raise util.AbstractMethodError

    @property
    def root(self) -> HTTP:
        """Get HTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def network_type(self):
        """Get network type for client."""
        raise util.AbstractMethodError

    @classmethod
    def create_from_http(cls: typing.Type[T], http) -> T:
        """
        Initialize HTTPBase directly from existing HTTP client.
        For internal use, do not use directly.

        :param http: HTTP client.
        """
        inst = cls.__new__(cls)
        inst._client = http._client
        inst._index = http._index
        inst._network_type = http._network_type
        return typing.cast(T, inst)

    def close(self):
        """Close the client session."""
        raise util.AbstractMethodError

    def __call__(self, cbs, *args, **kwds):
        """Invoke the NIS callback."""
        cb = cbs[self.index]
        # Force self.network_type to be executed on a different logical
        # block. Otherwise, we lead to infinite recursion when calling
        # get_network_type().
        try:
            network_type = kwds.pop('network_type')
        except KeyError:
            network_type = self.network_type

        return typing.cast(T, cb(self.raw, network_type, *args, **kwds))

    @property
    def _none(self):
        """Generate `None` with same evaluation as `network_type`."""
        raise util.AbstractMethodError


@util.inherit_doc
class HTTPBase(HTTPSharedBase):
    """Abstract base class for HTTP clients."""

    def __enter__(self) -> HTTPBase:
        raise util.AbstractMethodError

    def __exit__(self, exc_type, exc, tb) -> None:
        raise util.AbstractMethodError

    def close(self) -> None:
        self.raw.close()

    @property
    def raw(self) -> client.Client:
        try:
            return typing.cast(client.Client, self._client)
        except AttributeError:
            raise RuntimeError('Must be used inside `with` block.')

    @property
    def network_type(self) -> models.NetworkType:
        if self._network_type is None:
            network = self.root.network
            self._network_type = network.get_network_type()
        return self._network_type

    @property
    def _none(self) -> None:
        return None


@util.inherit_doc
class AsyncHTTPBase(HTTPSharedBase):
    """
    Abstract base class for asynchronous HTTP clients.

    :param endpoint: Domain name and port for the endpoint.
    :param loop: (Optional) Event loop for the client.
    """

    _loop: util.OptionalLoopType

    def __enter__(self) -> AsyncHTTPBase:
        raise TypeError("Only use async with.")

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    async def __aenter__(self) -> AsyncHTTPBase:
        raise util.AbstractMethodError

    async def __aexit__(self, exc_type, exc, tb) -> None:
        raise util.AbstractMethodError

    async def close(self) -> None:
        await self.raw.close()

    @property
    def raw(self) -> client.AsyncClient:
        try:
            return typing.cast(client.AsyncClient, self._client)
        except AttributeError:
            raise RuntimeError('Must be used inside `async with` block.')

    def __call__(self, cbs, *args, **kwds):
        return self.call(cbs, *args, **kwds)

    @util.observable
    async def call(self, cbs, *args, **kwds):
        return await super().__call__(cbs, *args, **kwds)

    @classmethod
    def create_from_http(cls: typing.Type[T], http) -> T:
        inst = super(AsyncHTTPBase, cls).create_from_http(http)  # type: ignore
        setattr(inst, '_loop', getattr(http, '_loop'))
        return inst  # type: ignore

    @property
    def loop(self) -> util.OptionalLoopType:
        """Get event loop."""
        return self._loop

    @property
    async def network_type(self) -> models.NetworkType:
        if self._network_type is None:
            network = self.root.network
            self._network_type = await network.get_network_type()
        return self._network_type

    @property
    async def _none(self) -> None:
        return None


class HTTP(HTTPSharedBase):
    """Abstract base class for the main HTTP client."""

    @property
    def account(self) -> AccountHTTP:
        """Get AccountHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def blockchain(self) -> BlockchainHTTP:
        """Get BlockchainHTTP to the same endpoint."""
        raise util.AbstractMethodError

    # @property
    # def contract(self) -> ContractHTTP:
    #    """Get ContractHTTP to the same endpoint."""
    #    raise util.AbstractMethodError

    @property
    def metadata(self) -> MetadataHTTP:
        """Get MetadataHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def config(self) -> ConfigHTTP:
        """Get ConfigHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def node(self) -> NodeHTTP:
        """Get NodeHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def mosaic(self) -> MosaicHTTP:
        """Get MosaicHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def namespace(self) -> NamespaceHTTP:
        """Get NamespaceHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def network(self) -> NetworkHTTP:
        """Get NetworkHTTP to the same endpoint."""
        raise util.AbstractMethodError

    @property
    def transaction(self) -> TransactionHTTP:
        """Get TransactionHTTP to the same endpoint."""
        raise util.AbstractMethodError


class AccountHTTP(HTTPSharedBase):
    """Abstract base class for the account HTTP client."""

    def get_account_info(
        self,
        address: models.Address,
        **kwds
    ):
        """
        Get info for account.

        :param address: Account address.
        :return: Account info object.
        """
        return self(nis.get_account_info, address, **kwds)

    def get_accounts_info(
        self,
        addresses: typing.Sequence[models.Address],
        **kwds
    ):
        """
        Get info for accounts.

        :param addresses: Sequence of account addresses.
        :return: List of account info objects.
        """
        return self(nis.get_accounts_info, addresses, **kwds)

    def get_account_properties(
        self,
        address: models.Address,
        **kwds
    ):
        """
        Get properties information for account.

        :param address: Account address.
        :return: AccountPropertiesInfo object.
        """
        return self(nis.get_account_properties, address, **kwds)

    def get_accounts_properties(
        self,
        addresses: typing.Sequence[models.Address],
        **kwds
    ):
        """
        Get properties information for accounts.

        :param addresses: Sequence of account addresses.
        :return: AccountPropertiesInfo object.
        """
        return self(nis.get_accounts_properties, addresses, **kwds)

    def get_multisig_account_info(
        self,
        address: models.Address,
        **kwds
    ):
        """
        Get multisig account information for account.

        :param address: Account address.
        :return: MultisigAccountInfo object.
        """
        return self(nis.get_multisig_account_info, address, **kwds)

    def get_multisig_account_graph_info(
        self,
        address: models.Address,
        **kwds
    ):
        """
        Get multisig account graph information for account.

        :param address: Account address.
        :return: MultisigAccountGraphInfo object.
        """
        return self(nis.get_multisig_account_graph_info, address, **kwds)

    def transactions(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get all transactions for account.

        :param public_account: Public key and address for account.
        :return: List of transaction objects.
        """
        return self(nis.get_account_transactions, public_account, **kwds)

    def incoming_transactions(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get all incoming transactions for account.

        :param public_account: Public key and address for account.
        :return: List of incoming transaction objects.
        """
        return self(nis.get_account_incoming_transactions, public_account, **kwds)

    def outgoing_transactions(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get all outgoing transactions for account.

        :param public_account: Public key and address for account.
        :return: List of outgoing transaction objects.
        """
        return self(nis.get_account_outgoing_transactions, public_account, **kwds)

    def unconfirmed_transactions(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get all unconfirmed transactions for account.

        :param public_account: Public key and address for account.
        :return: List of unconfirmed transaction objects.
        """
        return self(nis.get_account_unconfirmed_transactions, public_account, **kwds)

    def aggregate_bonded_transactions(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get all aggregate bonded transactions for account.

        :param public_account: Public key and address for account.
        :return: List of aggregate bonded transaction objects.
        """
        return self(nis.get_account_partial_transactions, public_account, **kwds)
# TODO: Check when stabilized
#     def contracts(
#         self,
#         public_account: models.PublicAccount,
#         **kwds
#     ):
#         """
#         Get account contracts.
#
#         :param public_account: Public account.
#         :return: List of ContractInfo object.
#         """
#         return self(nis.get_account_contracts, public_account, **kwds)

    def get_account_names(
        self,
        addresses: typing.Sequence[models.Address],
        **kwds
    ):
        """
        Get friendly names of accounts.

        :param addresses: Sequence of account addresses.
        :return: AccountNames object.
        """
        return self(nis.get_account_names, addresses, **kwds)


class BlockchainHTTP(HTTPSharedBase):
    """Abstract base class for the blockchain HTTP client."""

    def get_block_by_height(self, height: int, **kwds):
        """
        Get block information from the block height.

        :param height: Block height.
        :return: Information describing block.
        """
        return self(nis.get_block_by_height, height, **kwds)

    def get_blocks_by_height_with_limit(self, height: int, limit: int, **kwds):
        """
        Get information for blocks between [height, height+limit].

        :param height: Block height.
        :param limit: Maximum number of blocks to return.
        :return: Sequence of information models describing blocks.
        """
        return self(nis.get_blocks_by_height_with_limit, height, limit, **kwds)

    def get_block_transactions(self, height: int, **kwds):
        """
        Get information for all transactions included in a block by height.

        :param height: Block height.
        :return: Sequence of information models describing transactions.
        """
        return self(nis.get_block_transactions, height, **kwds)

    def get_merkle_by_hash_in_block(self, height: int, hash: str, **kwds):
        """
        Get information for all transactions included in a block by height.

        :param height: Block height.
        :param hash: Transaction hash.
        :return: Sequence of information models describing transactions.
        """
        return self(nis.get_merkle_by_hash_in_block, height, hash, **kwds)

    def get_block_receipts(self, height: int, **kwds):
        """
        Get receipts for a block by height.

        :param height: Block height.
        :return: Sequence of information models describing transactions.
        """
        return self(nis.get_block_receipts, height, **kwds)

    def get_blockchain_height(self, **kwds):
        """
        Get current blockchain height.

        :return: Blockchain height.
        """
        return self(nis.get_blockchain_height, **kwds)

    def get_blockchain_score(self, **kwds):
        """
        Get current blockchain score.

        :return: Blockchain score.
        """
        return self(nis.get_blockchain_score, **kwds)

    def get_diagnostic_blocks_by_height_with_limit(self, height: int, limit: int, **kwds):
        """
        Get diagnostic information for blocks between [height, height+limit].

        :param height: Block height.
        :param limit: Maximum number of blocks to return.
        :return: Sequence of information models describing blocks.
        """
        return self(nis.get_diagnostic_blocks_by_height_with_limit, height, limit, **kwds)

    def get_diagnostic_storage(self, **kwds):
        """
        Get diagnostic storage information for blockchain.

        :return: Blockchain diagnostic storage information.
        """
        return self(nis.get_diagnostic_storage, **kwds)

    def get_diagnostic_server(self, **kwds):
        """
        Get diagnostic storage information for rest server.

        :return: Blockchain diagnostic rest server information.
        """
        return self(nis.get_diagnostic_server, **kwds)


class ContractHTTP(HTTPSharedBase):
    """Abstract base class for the Contract HTTP client."""
    pass
# TODO: Check when stabilized
#     def get_contract(
#         self,
#         contract_id: str,
#         **kwds
#     ):
#         """
#         Gets the contract for a given contractId.
#
#         :param contractId: The account identifier.
#         :return: ContractInfo.
#         """
#         return self(nis.get_contract, contract_id, **kwds)
#
#     def get_contracts(
#         self,
#         addresses: typing.Sequence[models.Address],
#         **kwds
#     ):
#         """
#         Get contracts for an array of addresses
#
#         :param height: The height of the blockchain to get config.
#         :return: Sequence of ContractInfo.
#         """
#         return self(nis.get_contracts, addresses, **kwds)


class MetadataHTTP(HTTPSharedBase):
    """Abstract base class for the Metadata HTTP client."""

    def get_account_metadata(
        self,
        public_account: models.PublicAccount,
        **kwds
    ):
        """
        Get account metadata.

        :param public_account: Public account.
        :return: List of ContractInfo object.
        """
        return self(nis.get_account_metadata, public_account, **kwds)

    def get_mosaic_metadata(
        self,
        mosaic_id: models.MosaicId,
        **kwds
    ):
        """
        Get mosaic metadata.

        :param mosaic_id: Mosaic.
        :return: List of ContractInfo object.
        """
        return self(nis.get_mosaic_metadata, mosaic_id, **kwds)

    def get_namespace_metadata(
        self,
        namespace_id: models.NamespaceId,
        **kwds
    ):
        """
        Get namespace metadata.

        :param namespace_id: Namespace ID.
        :return: Metadata for a namespace.
        """
        return self(nis.get_namespace_metadata, namespace_id, **kwds)

    def get_metadata(
        self,
        metadata_id: str,
        **kwds
    ):
        """
        Gets the metadata for a given metadataId.

        :param metadataId: The metadata identifiers.
        :return: MetadataInfo.
        """
        return self(nis.get_metadata, metadata_id, **kwds)

    def get_metadatas(
        self,
        metadata_ids: typing.Sequence[str],
        **kwds
    ):
        """
        Gets the metadatas for a given metadataIds.

        :param metadataIds: The metadata identifiers.
        :return: Sequence of MetadataInfo.
        """
        return self(nis.get_metadatas, metadata_ids, **kwds)


class ConfigHTTP(HTTPSharedBase):
    """Abstract base class for the Config HTTP client."""

    def get_config(
        self,
        height: int,
        **kwds
    ):
        """
        Gets config of network at height.

        :param height: The height of the blockchain to get config.
        :return: CatapultConfig.
        """
        return self(nis.get_config, height, **kwds)

    def get_upgrade(
        self,
        height: int,
        **kwds
    ):
        """
        Gets upgrade of network at height.

        :param height: The height of the blockchain to get upgrade.
        :return: CatapultUpgrade.
        """
        return self(nis.get_upgrade, height, **kwds)


class NodeHTTP(HTTPSharedBase):
    """Abstract base class for the Node HTTP client."""

    def get_node_info(
        self,
        **kwds
    ):
        """
        Supplies additional information about the application running on a node.

        :return: NodeInfo.
        """
        return self(nis.get_node_info, **kwds)

    def get_node_time(
        self,
        **kwds
    ):
        """
        Gets the node time at the moment the reply was sent and received.

        :return: NodeTime.
        """
        return self(nis.get_node_time, **kwds)


class MosaicHTTP(HTTPSharedBase):
    """Abstract base class for the mosaic HTTP client."""

    def get_mosaic(
        self,
        id: models.MosaicId,
        **kwds
    ):
        """
        Get mosaic info from ID.

        :param id: Mosaic ID.
        :return: Mosaic info for ID.
        """
        return self(nis.get_mosaic, id, **kwds)

    def get_mosaics(
        self,
        ids: typing.Sequence[models.MosaicId],
        **kwds
    ):
        """
        Get mosaic info from IDs.

        :param ids: Sequence of mosaic IDs.
        :return: Mosaic info list for IDS.
        """
        return self(nis.get_mosaics, ids, **kwds)

    def get_mosaic_names(
        self,
        ids: typing.Sequence[models.MosaicId],
        **kwds
    ):
        """
        Get mosaic names from IDs.

        :param ids: Sequence of mosaic IDs.
        :return: Mosaic names for IDS.
        """
        return self(nis.get_mosaic_names, ids, **kwds)

    def get_mosaic_richlist(
        self,
        mosaic_id: models.MosaicId,
        **kwds
    ):
        """
        Get account balances in a given Mosaic.

        :param mosaic_id: Mosaic ID.
        :return: Account balances.
        """
        return self(nis.get_mosaic_richlist, mosaic_id, **kwds)


class NamespaceHTTP(HTTPSharedBase):
    """Abstract base class for the namespace HTTP client."""

    def get_namespace(
        self,
        namespace_id: models.NamespaceId,
        **kwds
    ):
        """
        Get namespace information from ID.

        :param id: Namespace ID.
        :return: Namespace information.
        """
        return self(nis.get_namespace, namespace_id, **kwds)

    def get_namespaces_from_account(
        self,
        address: models.Address,
        **kwds
    ):
        """
        Get namespaces owned by account.

        :param address: Account address.
        :return: List of namespace information objects.
        """
        return self(nis.get_namespaces_from_account, address, **kwds)

    def get_namespaces_from_accounts(
        self,
        addresses: typing.Sequence[models.Address],
        **kwds
    ):
        """
        Get namespaces owned by accounts.

        :param addresses: Sequence of account addresses.
        :return: List of namespace information objects.
        """
        return self(nis.get_namespaces_from_accounts, addresses, **kwds)

    def get_namespaces_name(
        self,
        ids: typing.Sequence[models.NamespaceId],
        **kwds
    ):
        """
        Get namespace names from IDs.

        :param ids: Sequence of namespace IDs.
        :return: Namespace names for IDS.
        """
        return self(nis.get_namespaces_name, ids, **kwds)

    def get_linked_mosaic_id(
        self,
        namespace_id: models.NamespaceId,
        **kwds
    ):
        """
        Get mosaic ID from linked mosaic alias.

        :param id: Namespace ID.
        :return: Mosaic ID.
        """
        return self(nis.get_linked_mosaic_id, namespace_id, **kwds)

    def get_linked_address(
        self,
        namespace_id: models.NamespaceId,
        **kwds
    ):
        """
        Get address from linked address alias.

        :param id: Namespace ID.
        :return: Address object.
        """
        return self(nis.get_linked_address, namespace_id, **kwds)


class NetworkHTTP(HTTPSharedBase):
    """Abstract base class for the network HTTP client."""

    def get_network_type(self, **kwds):
        """
        Get network type.

        :return: Network type.
        """
        return self(nis.get_network_type, network_type=self._none, **kwds)


class TransactionHTTP(HTTPSharedBase):
    """Abstract base class for the transaction HTTP client."""

    def get_transaction(self, hash: str, **kwds):
        """
        Get transaction info by hash.

        :param hash: Transaction hash.
        :return: Transaction info.
        """
        return self(nis.get_transaction, hash, **kwds)

    def get_transactions(self, hashes: typing.Sequence[str], **kwds):
        """
        Get transaction info by hash.

        :param hashes: Sequence of transaction hashes.
        :return: Transaction info.
        """
        return self(nis.get_transactions, hashes, **kwds)

    def get_transaction_status(self, hash: str, **kwds):
        """
        Get transaction status by hash.

        :param hash: Transaction hash.
        :return: Transaction status.
        """
        return self(nis.get_transaction_status, hash, **kwds)

    def get_transaction_statuses(self, hashes: typing.Sequence[str], **kwds):
        """
        Get transaction status by sequence of hashes.

        :param hashes: Sequence of transaction hashes.
        :return: List of transaction statuses.
        """
        return self(nis.get_transaction_statuses, hashes, **kwds)

    # TODO(ahuszagh)
    # getTransactionEffectiveFee

    def announce(
        self,
        transaction: models.SignedTransaction,
        **kwds
    ):
        """
        Announce transaction to network.

        :param transaction: Signed transaction data.
        :return: Transaction announce response.
        """
        return self(nis.announce, transaction, **kwds)

    def announce_partial(
        self,
        transaction: models.SignedTransaction,
        **kwds
    ):
        """
        Announce partial transaction to network.

        :param transaction: Signed transaction data.
        :return: Transaction announce response.
        """
        return self(nis.announce_partial, transaction, **kwds)

    def announce_cosignature(
        self,
        transaction: models.SignedTransaction,
        **kwds
    ):
        """
        Announce cosignature transaction to network.

        :param transaction: Signed transaction data.
        :return: Transaction announce response.
        """
        return self(nis.announce_cosignature, transaction, **kwds)

# WEBSOCKET
# ---------


@util.dataclass(frozen=True)
class ListenerMessage(util.Object):
    """Message from a listener."""

    channel_name: str
    message: MessageType


@util.observable
class Listener(util.Object):
    """
    Abstract base class for the websockets-based listener.

    :param endpoint: Domain name and port for the endpoint.
    """

    _client: client.WebsocketClient
    _iter_: typing.AsyncIterator[bytes]
    _conn: typing.AsyncContextManager
    _loop: util.OptionalLoopType
    _uid: typing.Optional[str] = None

    def __enter__(self) -> Listener:
        raise TypeError("Only use async with.")

    def __exit__(self, exc_type, exc, tb) -> None:
        pass

    async def __aenter__(self):
        raise util.AbstractMethodError

    async def __aexit__(self, exc_type, exc, tb) -> None:
        raise util.AbstractMethodError

    @property
    def raw(self) -> client.WebsocketClient:
        try:
            return self._client
        except AttributeError:
            raise RuntimeError('Must be used inside `async with` block.')

    @property
    def _iter(self) -> typing.AsyncIterator[bytes]:
        try:
            return self._iter_
        except AttributeError:
            raise RuntimeError('Must be used inside `async with` block.')

    @property
    def closed(self) -> bool:
        """Get if client session has been closed."""
        return self.raw.closed

    @property
    async def uid(self) -> str:
        """Get UUID (unique identifier) for WS requests."""

        if self._uid is None:
            self._uid = json.loads(await self.raw.recv())['uid']
        return self._uid

    def __aiter__(self) -> Listener:
        return self

    async def __anext__(self) -> typing.Optional[ListenerMessage]:
        """Iterate over subscribed messages."""

        message: bytes = await self._iter.__anext__()
        data = json.loads(message)
        if 'transaction' in data:
            # New transaction data.
            channel_name = typing.cast(str, data['meta'].pop('channelName'))
            transaction = models.Transaction.create_from_dto(data)
            return ListenerMessage(channel_name, transaction)
        elif 'block' in data:
            # New block info.
            block = models.BlockInfo.create_from_dto(data)
            return ListenerMessage('block', block)
        elif 'status' in data:
            # New transaction status error.
            error = models.TransactionStatusError.create_from_dto(data)
            return ListenerMessage('status', error)
        elif 'meta' in data:
            # New metadata.
            channel_name = typing.cast(str, data['meta']['channelName'])
            hash = typing.cast(str, data['meta']['hash'])
            return ListenerMessage(channel_name, hash)
        elif 'parentHash' in data:
            # New cosignature for transaction.
            cosignature = models.CosignatureSignedTransaction.create_from_dto(data)
            return ListenerMessage('cosignature', cosignature)
        else:
            # Unknown data information, don't pollute the message,
            # only send the information's keys.
            msg = f"Unknown message from Listener subscription, keys are {data.keys()}."
            raise ValueError(msg)

    @util.observable
    async def new_block(self) -> None:
        """Emit message when new blocks are added to chain."""
        await self.subscribe('block')

    @util.observable
    async def confirmed(self, address: models.Address) -> None:
        """Emit message when new transactions are confirmed for a given address."""
        await self.subscribe(f'confirmedAdded/{address.address}')

    @util.observable
    async def confirmed_added(self, address: models.Address) -> None:
        """Emit message when new transactions are confirmed for a given address."""
        await self.confirmed(address)

    @util.observable
    async def unconfirmed_added(self, address: models.Address) -> None:
        """Emit message for unconfirmed transactions are announced."""
        await self.subscribe(f'unconfirmedAdded/{address.address}')

    @util.observable
    async def unconfirmed_removed(self, address: models.Address) -> None:
        """Emit message when unconfirmed transactions change state."""
        await self.subscribe(f'unconfirmedRemoved/{address.address}')

    @util.observable
    async def aggregate_bonded_added(self, address: models.Address) -> None:
        """Emit message when new, unconfirmed, aggregate transactions are announced."""
        await self.subscribe(f'partialAdded/{address.address}')

    @util.observable
    async def aggregate_bonded_removed(self, address: models.Address) -> None:
        """Emit message when unconfirmed, aggregate transactions change state."""
        await self.subscribe(f'partialRemoved/{address.address}')

    @util.observable
    async def status(self, address: models.Address) -> None:
        """Emit message each time a transaction contains an error."""
        await self.subscribe(f'status/{address.address}')

    @util.observable
    async def cosignature_added(self, address: models.Address) -> None:
        """Emit message each time a cosigner signs a message"""
        await self.subscribe(f'cosignature/{address.address}')

    async def subscribe(self, channel: str) -> None:
        """Subscribe to websockets channel."""

        message = json.dumps({
            'uid': await self.uid,
            'subscribe': channel
        })
        await self.raw.send(message)

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from websockets channel."""

        message = json.dumps({
            'uid': await self.uid,
            'unsubscribe': channel
        })
        await self.raw.send(message)



"""
    from_async_iterable
    ===================

    RxPy extension to generate an observable from an async iterable.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import asyncio
import rx
from rx.internal import extensionclassmethod


def from_async_iterable(cls, asyncgen, scheduler=None, loop=None):      # noqa: C901
    """
    Generate observable from an asynchronous iterable.

    :param asyncgen: Asynchronous iterable to wrap as observable.
    :param scheduler: (Optional) scheduler to determine the event loop.
    :param loop: (Optional) Event loop. Has higher precedence than scheduler.
    """

    # Get the current event loop.
    if loop is None:
        loop = asyncio.get_event_loop()
    elif scheduler is not None:
        loop = scheduler.loop

    def subscribe(observer):
        async def worker(aiter):
            """Worker to process all the items and create a coroutine."""

            # Only call __aiter__ if the type doesn't implement
            # __anext__, to avoid infinite recursion.
            if not hasattr(aiter, '__anext__'):
                aiter = aiter.__aiter__()

            # Simulate an `async for` loop, using `__anext__`. This
            # is to avoid recursively calling `__aiter__`, which is
            # what happens when we wrap `__aiter__` into an `Observable`.
            try:
                while True:
                    observer.on_next(await aiter.__anext__())
            except StopAsyncIteration:
                pass

        def done(future):
            """Callback for when the worker is done."""

            try:
                future.result()
                observer.on_completed()
            except Exception as ex:
                observer.on_error(ex)

        # Wrap worker into a task and signal complete when done.
        task = loop.create_task(worker(asyncgen))
        task.add_done_callback(done)

        def dispose():
            """Allow the task to be cancelable."""

            if task and task.cancel:
                task.cancel()

        return dispose

    return rx.AnonymousObservable(subscribe)


extensionclassmethod(rx.AnonymousObservable)(from_async_iterable)
extensionclassmethod(rx.Observable)(from_async_iterable)



"""
    blockchain_upgrade_transaction
    ====================

    Transfer transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = [
    'BlockchainUpgradeTransaction',
    'BlockchainUpgradeInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('BLOCKCHAIN_UPGRADE')
class BlockchainUpgradeTransaction(Transaction):
    """
    Blockchain upgrade transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param new_blockchain_version: New blockchain version
    :param upgrade_period: Upgrade period
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    new_blockchain_version: int
    upgrade_period: int

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        new_blockchain_version: int,
        upgrade_period: int,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None
    ) -> None:
        super().__init__(
            TransactionType.TRANSFER,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info
        )
        self._set('new_blockchain_version', new_blockchain_version)
        self._set('upgrade_period', new_blockchain_version)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        network_type: NetworkType,
        new_blockchain_version: int,
        upgrade_period: int,
        max_fee: int = 0,
    ):
        """
        Create new network configuration transaction.

        :param deadline: Deadline to include transaction.
        :param network_type: Network type.
        :param new_blockchain_version: New blockchain version
        :param upgrade_period: Upgrade period
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.TRANSFER,
            deadline,
            new_blockchain_version,
            upgrade_period,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        new_blockchain_version_size = util.U64_BYTES
        upgrade_period_size = util.U64_BYTES

        return new_blockchain_version_size + upgrade_period_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export transfer-specific data to catbuffer."""

        # uint64_t upgrade_period
        # uint64_t new_blockchain_version
        upgrade_period = util.u64_to_catbuffer(self.upgrade_period)
        new_blockchain_version = util.u64_to_catbuffer(self.new_blockchain_version)

        return upgrade_period + new_blockchain_version

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transfer-specific data from catbuffer."""

        # uint64_t upgrade_period
        # uint64_t new_blockchain_version
        upgrade_period = util.u64_from_catbuffer(data[:util.U64_BYTES])
        data = data[util.U64_BYTES:]
        new_blockchain_version = util.u64_from_catbuffer(data[:util.U64_BYTES])
        data = data[util.U64_BYTES:]

        self._set('upgrade_period', upgrade_period)
        self._set('new_blockchain_version', new_blockchain_version)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'newBlockchainVersion', 'upgradePeriod'}

        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'newBlockchainVersion': util.u64_to_dto(self.new_blockchain_version),
            'upgradePeriod': util.u64_to_dto(self.upgrade_period),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        new_blockchain_version = util.u64_from_dto(data['newBlockchainVersion'])
        upgrade_period = util.u64_from_dto(data['upgradePeriod'])

        self._set('new_blockchain_version', new_blockchain_version)
        self._set('upgrade_period', upgrade_period)


@register_transaction('BLOCKCHAIN_UPGRADE')
class BlockchainUpgradeInnerTransaction(InnerTransaction, BlockchainUpgradeTransaction):
    """Embedded blockchain upgrade transaction."""

    __slots__ = ()
